/*********************************************************************
*                SEGGER MICROCONTROLLER GmbH & Co. KG                *
*        Solutions for real time microcontroller applications        *
**********************************************************************
*                                                                    *
*        (c) 2003-2010     SEGGER Microcontroller GmbH & Co KG       *
*                                                                    *
*        Internet: www.segger.com    Support:  support@segger.com    *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File        : LCDConf.c
Purpose     : Display controller configuration (single layer)
---------------------------END-OF-HEADER------------------------------
*/
#include <stdlib.h>   // for rand() function

#include "lpclib.h"

#include "bsp-hitex4300.h"

#include "GUI.h"
#include "GUIDRV_Lin.h"


/* Physical display size */
#define XSIZE_PHYS 240
#define YSIZE_PHYS 320

/* Virtual display size */
#define VXSIZE_PHYS (XSIZE_PHYS)
#define VYSIZE_PHYS (YSIZE_PHYS)

/* Video RAM address */
#define VRAM_ADDR_PHYS  (U32)&_aVRAM[0]



static const LCD_Config tftTimings = {
    .opcode = LCD_OPCODE_SET_TIMINGS,
    {.timings = {
        .sizeX = XSIZE_PHYS,
        .sizeY = YSIZE_PHYS,
        .frontPorchH = 10,
        .syncWidthH = 2,
        .backPorchH = 28,
        .frontPorchV = 1,
        .syncWidthV = 2,
        .backPorchV = 2,
        .polarityHSYNC = LCD_POLARITY_ACTIVE_LOW,
        .polarityVSYNC = LCD_POLARITY_ACTIVE_LOW,
        .polarityClock = LCD_POLARITY_INVERTED,
        .polarityENAB = LCD_POLARITY_ACTIVE_HIGH,
        .bitsPerPixel = LCD_BPP_16_565, }},
};


static const LCD_Config tftEnable = {
    .opcode = LCD_OPCODE_SET_POWER,
    {.powered = ENABLE, },
};


LCD_Handle tft;



/* Color conversion */
#define COLOR_CONVERSION  GUICC_M565

/* Pixel width in bytes */
#define PIXEL_WIDTH_NOM     2
#define PIXEL_WIDTH_DENOM   1

/* Display driver */
#define DISPLAY_DRIVER  &GUIDRV_Lin_OSX_16_API


#define TFT_DATA()         do { GPIO_writeBit(TFT_CMD_DATA, 1); } while(0)
#define TFT_CMD()          do { GPIO_writeBit(TFT_CMD_DATA, 0); } while(0)


/* Touch controller settings */
#define TOUCH_AD_RIGHT        275
#define TOUCH_AD_LEFT         3845
#define TOUCH_AD_TOP          3873
#define TOUCH_AD_BOTTOM       215

#define ADS_START                      (1 << 7)
#define ADS_A2A1A0_d_y                 (1 << 4)  // Differential
#define ADS_A2A1A0_d_z1                (3 << 4)  // Differential
#define ADS_A2A1A0_d_z2                (4 << 4)  // Differential
#define ADS_A2A1A0_d_x                 (5 << 4)  // Differential
#define ADS_A2A1A0_vaux                (6 << 4)  // Non-differential
#define ADS_12_BIT                     (0 << 3)
#define ADS_SER                        (1 << 2)  // Non-differential
#define ADS_DFR                        (0 << 2)  // Differential
#define ADS_PD10_ADC_ON                (1 << 0)  // ADC on
#define ADS_PD10_REF_ON                (2 << 0)  // vREF on + penirq
#define ADS_PD10_ALL_ON                (3 << 0)  // ADC + vREF on

#define READ_12BIT_DFR(D, ADC, VREF)  (ADS_START                    | \
                                       D                            | \
                                       ADS_12_BIT                   | \
                                       ADS_DFR                      | \
                                       (ADC  ? ADS_PD10_ADC_ON : 0) | \
                                       (VREF ? ADS_PD10_REF_ON : 0))


#define READ_12BIT_SER(x)              (ADS_START  | /* Single-ended samples need to power   */ \
                                        x          | /* up reference voltage first; therefor */ \
                                        ADS_12_BIT | /* we leave both VREF and ADC powered.  */ \
                                        ADS_SER)

#define REF_ON                         (READ_12BIT_DFR(ADS_A2A1A0_d_x,  1, 1))
#define READ_Y(VREF)                   (READ_12BIT_DFR(ADS_A2A1A0_d_y,  1, VREF))
#define READ_Z1(VREF)                  (READ_12BIT_DFR(ADS_A2A1A0_d_z1, 1, VREF))
#define READ_Z2(VREF)                  (READ_12BIT_DFR(ADS_A2A1A0_d_z2, 1, VREF))
#define READ_X(VREF)                   (READ_12BIT_DFR(ADS_A2A1A0_d_x,  1, VREF))
#define PWRDOWN                        (READ_12BIT_DFR(ADS_A2A1A0_d_y,  0, 0))

#define TS_DEBOUNCE_MAX                10
#define TS_DEBOUNCE_TOL                3

#ifndef   ABS
  #define ABS(x)                       (((int32_t)(x)) < 0 ? (-x) : (x))
#endif


/*********************************************************************
*
*       Configuration checking
*
**********************************************************************
*/
#ifndef   VRAM_ADDR_VIRT
  #define VRAM_ADDR_VIRT  0
#endif



/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/* Video RAM */
#ifdef __ICCARM__
    #pragma data_alignment=8  // 8 byte align frame buffer to be on the safe side (some display controllers need this)
    #pragma location="VRAM"
    static __no_init U32 _aVRAM[VXSIZE_PHYS * VYSIZE_PHYS / (4 / PIXEL_WIDTH)];
#endif
#ifdef __CC_ARM
    U32 static _aVRAM[VXSIZE_PHYS * VYSIZE_PHYS / ((4 * PIXEL_WIDTH_DENOM) / PIXEL_WIDTH_NOM)] __attribute__ ((section ("VRAM"), zero_init));
#endif
#ifdef __GNUC__
    static U32 _aVRAM[VXSIZE_PHYS * VYSIZE_PHYS / ((4 * PIXEL_WIDTH_DENOM) / PIXEL_WIDTH_NOM)] __attribute__ ((section ("VRAM")));
#endif


#if GUI_SUPPORT_TOUCH  // Used when touch screen support is enabled
static int _TouchX;
static int _TouchY;
static U8  _PenIsDown;
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/


/*********************************************************************
*
*       _SetDisplayOrigin()
*/
static void _SetDisplayOrigin (int x, int y)
{
    (void)  x;

    /* Set start address for display data and enable LCD controller */
    LPC_LCD->UPBASE = VRAM_ADDR_PHYS + (y * YSIZE_PHYS * PIXEL_WIDTH_NOM) / PIXEL_WIDTH_DENOM;  // Needs to be set, before LCDC is enabled
}



static const SSP_Config tftSpiFormat[] = {
    {.opcode = SSP_OPCODE_SET_FORMAT, .continued = ENABLE,
        {.format = {
            .bits = SSP_BITS_8,
            .clockFormat = SSP_CLOCK_CPHA1_CPOL1,
            .frameFormat = SSP_FRAMEFORMAT_SPI, }}},
    {.opcode = SSP_OPCODE_SET_BITRATE, .continued = DISABLE,
        {.bitrate = 1000000ul, }},
};



void _SSP_Send (uint8_t *data, int numBytes)
{
    uint8_t dummy[20];  //TODO....

    GENERICSPI_sendAndReceive(tftController, &tftSpiFormat[0], data, dummy, numBytes);
}




/** Write to an LCD register. */
static void _WriteLcdReg (uint16_t address, uint16_t data)
{
    uint8_t buffer[2];

    TFT_CMD();
    buffer[0] = 0;
    buffer[1] = address & 0xFF;
    _SSP_Send(buffer, 2);

    TFT_DATA();
    buffer[0] = data >> 8;
    buffer[1] = data & 0xFF;
    _SSP_Send(buffer, 2);

    TFT_CMD();
    buffer[0] = 0;
    buffer[1] = 0x22;
    _SSP_Send(buffer, 2);
}



/** Init controller ILI9322 on TFT display. */
static void _InitLcdController (void)
{
    /* Display init sequence */
  _WriteLcdReg(0x00,0x0001);
    LPCLIB_OSALTaskDelay(NULL, NULL, 30);
  _WriteLcdReg(0x03,0x6E3E);  // AAAC
  _WriteLcdReg(0x0C,0x0007);  // 0002
  _WriteLcdReg(0x0D,0x000E);  // 000A
  _WriteLcdReg(0x0E,0x2C00);  // 2C00
  _WriteLcdReg(0x1E,0x00AE);  // 00B8
    LPCLIB_OSALTaskDelay(NULL, NULL, 30);

  _WriteLcdReg(0x07,0x0021);
    LPCLIB_OSALTaskDelay(NULL, NULL, 100);
  _WriteLcdReg(0x07,0x0023);
    LPCLIB_OSALTaskDelay(NULL, NULL, 100);
  _WriteLcdReg(0x07,0x0033);
    LPCLIB_OSALTaskDelay(NULL, NULL, 100);

  _WriteLcdReg(0x01,0x2B3F);
  _WriteLcdReg(0x02,0x0600);
  _WriteLcdReg(0x10,0x0000);
    LPCLIB_OSALTaskDelay(NULL, NULL, 30);
  _WriteLcdReg(0x11,0xC5B0);  // 60B0: RGB I/R
    LPCLIB_OSALTaskDelay(NULL, NULL, 40);
  _WriteLcdReg(0x15,0x00D0);

  _WriteLcdReg(0x05,0x0000);
  _WriteLcdReg(0x06,0x0000);
  _WriteLcdReg(0x16,0xEF1C);
  _WriteLcdReg(0x17,0x0003);
  _WriteLcdReg(0x07,0x0233);
  _WriteLcdReg(0x0B,0x5310);
  _WriteLcdReg(0x0F,0x0000);
  _WriteLcdReg(0x25,0xE000);
    LPCLIB_OSALTaskDelay(NULL, NULL, 40);

  _WriteLcdReg(0x41,0x0000);
  _WriteLcdReg(0x42,0x0000);
  _WriteLcdReg(0x48,0x0000);
  _WriteLcdReg(0x49,0x013F);
  _WriteLcdReg(0x44,0xEF00);
  _WriteLcdReg(0x45,0x0000);
  _WriteLcdReg(0x46,0x013F);
  _WriteLcdReg(0x4A,0x0000);
  _WriteLcdReg(0x4B,0x0000);
    LPCLIB_OSALTaskDelay(NULL, NULL, 40);

  _WriteLcdReg(0x30,0x0707);
  _WriteLcdReg(0x31,0x0600);  // 0704
  _WriteLcdReg(0x32,0x0005);  // 0204
  _WriteLcdReg(0x33,0x0402);  // 0201
  _WriteLcdReg(0x34,0x0203);
  _WriteLcdReg(0x35,0x0204);
  _WriteLcdReg(0x36,0x0204);
  _WriteLcdReg(0x37,0x0401);  // 0502
  _WriteLcdReg(0x3A,0x0302);
  _WriteLcdReg(0x3B,0x0500);
    LPCLIB_OSALTaskDelay(NULL, NULL, 40);
  _WriteLcdReg(0x22,0x0000);

    LPCLIB_OSALTaskDelay(NULL, NULL, 40);
}


/*********************************************************************
*
*       _InitLCD
*
* Function description:
*   Sets LCD configuration and initializes the LCD.
*/
static void _InitLCD (unsigned LayerIndex)
{
int i;
for(i=0;i<sizeof(_aVRAM)/4;i++) {
  _aVRAM[i] = rand();
}

    LCD_open(LCD0, &tft);

    /* emWin: Set display size and video-RAM address */
    LCD_SetSizeEx (XSIZE_PHYS, YSIZE_PHYS, LayerIndex);
    LCD_SetVSizeEx(VXSIZE_PHYS, VYSIZE_PHYS, LayerIndex);
    LCD_SetVRAMAddrEx(LayerIndex, (void*)VRAM_ADDR_PHYS);

    /* Init LCD */
    GPIO_writeBit(TFT_RESET, 1);
    LCD_ioctl(tft, &tftTimings);
    _SetDisplayOrigin(0, 0);

    _InitLcdController();

    LCD_ioctl(tft, &tftEnable);
}



#if GUI_SUPPORT_TOUCH  // Used when touch screen support is enabled


/*********************************************************************
*
*       _CheckUpdateTouch
*
* Function description:
*   Reads touch values from ADC on Truly LCD display
*   controller IC on Embedded Artists QVGA Base Board.
*   Checks if a touch event has occurred. If we found an event the
*   static variables for x and y will be filled with new values that
*   can be processed via GUI touch routines.
*
* Return value:
*      0  No touch event, x, y have not been updated
*   != 0  Touch event occurred, x, y have been updated and return value is read z value
*/
static int _CheckUpdateTouch(void) {
    uint16_t x;
    uint16_t y;
  int z;
    uint16_t z1;
    uint16_t z2;
    uint16_t LastX;
    uint16_t LastY;
    uint16_t LastZ1;
    uint16_t LastZ2;
  int i;

    TSC2046_sendCommand(touchController, READ_Y(0), &LastY);
    LastY >>= 3;

    TSC2046_sendCommand(touchController, READ_X(0), &LastX);
    LastX >>= 3;

    TSC2046_sendCommand(touchController, READ_Z1(0), &LastZ1);
    LastZ1 >>= 3;

    TSC2046_sendCommand(touchController, READ_Z2(0), &LastZ2);
    LastZ2 >>= 3;

    /* Check for touch event */
    z = (LastX * LastZ2 - LastZ1) / LastZ1;

    if ((z > 40000) || (z == 0)) {
        return 0;  /* No update */
    }

    /* Get debounced y value */
    for (i = 0; i < TS_DEBOUNCE_MAX; i++) {
        TSC2046_sendCommand(touchController, READ_Y(0), &y);
        y >>= 3;
        if (ABS(LastY - y) <= TS_DEBOUNCE_TOL) {
            break;
        }
        LastY = y;
    }

    /* Get debounced x value */
    for (i = 0; i < TS_DEBOUNCE_MAX; i++) {
        TSC2046_sendCommand(touchController, READ_X(0), &x);
        x >>= 3;
        if (ABS(LastX - x) <= TS_DEBOUNCE_TOL) {
            break;
        }
        LastX = x;
    }

    /* Get debounced z1 value */
    for (i = 0; i < TS_DEBOUNCE_MAX; i++) {
        TSC2046_sendCommand(touchController, READ_Z1(0), &z1);
        z1 >>= 3;
        if (ABS(LastZ1 - z1) <= TS_DEBOUNCE_TOL) {
            break;
        }
        LastZ1 = z1;
    }

    /* Get debounced z2 value */
    for (i = 0; i < TS_DEBOUNCE_MAX; i++) {
        TSC2046_sendCommand(touchController, READ_Z2(0), &z2);
        z2 >>= 3;
        if (ABS(LastZ2 - z2) <= TS_DEBOUNCE_TOL) {
            break;
        }
        LastZ2 = z2;
    }

    /* Update static variables if we have a valid touch */
    z = ((x * z2) - z1) / z1;
    if (z > 40000) {
        return 0;  // No update
    }
    _TouchX = y;  // Touch is swapped
    _TouchY = x;  // Touch is swapped
    return z;
}

/*********************************************************************
*
*       _ExecTouch
*
* Function description:
*   Check for new touch event. Static x, y coordinates will be updated
*   by the _CheckUpdateTouch() routine. If no touch event has occurred
*   we do nothing.
*/
void _ExecTouch(void) {
  int HasUpdate;
  GUI_PID_STATE State;
    uint16_t dummy16;


  HasUpdate = _CheckUpdateTouch();

    TSC2046_sendCommand(touchController, PWRDOWN, &dummy16);
  if (HasUpdate) {
    _PenIsDown = 1;
    GUI_TOUCH_Exec();
  } else if (_PenIsDown) {
    //
    // No further touch event after at least one touch event means we have
    // lift the pen from the touch screen which means a click.
    //
    _PenIsDown = 0;
    GUI_PID_GetState(&State);
    State.Pressed = 0;
    GUI_PID_StoreState(&State);
  }
  // OS_RetriggerTimer(&_TouchTimer);
}

/*********************************************************************
*
*       _InitTouch
*
* Function description:
*   Initializes the touch screen.
*/
static void _InitTouch(void) {
    U32 TouchOrientation;
    U8  Config[3];


  //
  // Init ports and SSP interface as needed for touch
  //
  Config[0] = REF_ON;
  Config[1] = (READ_12BIT_SER(ADS_A2A1A0_vaux) | ADS_PD10_ALL_ON);
  Config[2] = PWRDOWN;
    TSC2046_sendCommandSequence(touchController, Config, 3);


  //
  // Calibrate touch
  //
  TouchOrientation = (GUI_MIRROR_X * LCD_GetMirrorXEx(0)) |
                     (GUI_MIRROR_Y * LCD_GetMirrorYEx(0)) |
                     (GUI_SWAP_XY  * LCD_GetSwapXYEx (0)) ;

  GUI_TOUCH_SetOrientation(TouchOrientation);
  if (LCD_GetSwapXYEx(0)) {
    GUI_TOUCH_Calibrate(GUI_COORD_X, 0, XSIZE_PHYS, TOUCH_AD_LEFT, TOUCH_AD_RIGHT);   // x axis swapped
    GUI_TOUCH_Calibrate(GUI_COORD_Y, 0, YSIZE_PHYS, TOUCH_AD_TOP , TOUCH_AD_BOTTOM);  // y axis swapped
  } else {
    GUI_TOUCH_Calibrate(GUI_COORD_X, 0, YSIZE_PHYS, TOUCH_AD_LEFT, TOUCH_AD_RIGHT);   // x axis
    GUI_TOUCH_Calibrate(GUI_COORD_Y, 0, XSIZE_PHYS, TOUCH_AD_TOP , TOUCH_AD_BOTTOM);  // y axis
  }
}


#endif  // GUI_SUPPORT_TOUCH

/*********************************************************************
*
*       _InitController
*
* Function description:
*   Initializes the LCD controller and touch screen
*
*/
static void _InitController (unsigned LayerIndex)
{
    _InitLCD(LayerIndex);
#if GUI_SUPPORT_TOUCH  // Used when touch screen support is enabled
    _InitTouch();
#endif
}


/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/** Set up the display driver configuration.
 *
 *  Called by emWIN library.
 */
void LCD_X_Config (void)
{
    /* Set display driver and color conversion for 1st layer */
    GUI_DEVICE_CreateAndLink(DISPLAY_DRIVER, COLOR_CONVERSION, 0, 0);

    /* Display driver configuration, required for Lin-driver */
    LCD_SetPosEx     (0, 0, 0);
    if (LCD_GetSwapXYEx(0)) {
        LCD_SetSizeEx  (0, YSIZE_PHYS , XSIZE_PHYS);
        LCD_SetVSizeEx (0, VYSIZE_PHYS, VXSIZE_PHYS);
    }
    else {
        LCD_SetSizeEx  (0, XSIZE_PHYS , YSIZE_PHYS);
        LCD_SetVSizeEx (0, VXSIZE_PHYS, VYSIZE_PHYS);
    }
    LCD_SetVRAMAddrEx(0, (void*)VRAM_ADDR_VIRT);

    /* Set user palette data (only required if no fixed palette is used) */
  #if defined(PALETTE)
    LCD_SetLUTEx(0, PALETTE);
  #endif
}


/** Execute emWIN commands at run-time
 *
 *  This function is called by the display driver for several purposes.
 *  To support the according task the routine needs to be adapted to
 *  the display controller. Please note that the commands marked with
 *  'optional' are not cogently required and should only be adapted if
 *  the display controller supports these features.
 *
 *  \param[in] LayerIndex Index of layer to be configured
 *  \param[in] Cmd Please refer to the details in the switch statement below
 *  \param pData Pointer to a LCD_X_DATA structure
 *  \retval <-1 Error
 *  \retval -1 Command not handled
 *  \retval 0 Ok
 */
int LCD_X_DisplayDriver (unsigned LayerIndex, unsigned Cmd, void * pData)
{
    int r;
    LCD_X_SETORG_INFO      * pSetOrg;

    (void) LayerIndex;


    switch (Cmd) {
    case LCD_X_INITCONTROLLER:              /* Required! */
        //
        // Called during the initialization process in order to set up the
        // display controller and put it into operation. If the display
        // controller is not initialized by any external routine this needs
        // to be adapted by the customer...
        //
        // ...
        _InitController(0);
        return 0;

    case LCD_X_SETORG:
        //
        // Required for setting the display origin which is passed in the 'xPos' and 'yPos' element of p
        //
        pSetOrg = (LCD_X_SETORG_INFO *)pData;
        _SetDisplayOrigin(pSetOrg->xPos, pSetOrg->yPos);
        return 0;
    
    default:
        r = -1;
        break;
    }

    return r;
}


#if GUI_SUPPORT_TOUCH  // Used when touch screen support is enabled

/*********************************************************************
*
*       GUI_TOUCH_X_ActivateX()
*
* Function decription:
*   Called from GUI, if touch support is enabled.
*   Switches on voltage on X-axis,
*   prepares measurement for Y-axis.
*   Voltage on Y-axis is switched off.
*/
void GUI_TOUCH_X_ActivateX(void) {
}

/*********************************************************************
*
*       GUI_TOUCH_X_ActivateY()
*
* Function decription:
*   Called from GUI, if touch support is enabled.
*   Switches on voltage on Y-axis,
*   prepares measurement for X-axis.
*   Voltage on X-axis is switched off.
*/
void GUI_TOUCH_X_ActivateY(void) {
}

/*********************************************************************
*
*       GUI_TOUCH_X_MeasureX()
*
* Function decription:
*   Called from GUI, if touch support is enabled.
*   Measures voltage of X-axis.
*/
int  GUI_TOUCH_X_MeasureX(void) {
  return _TouchX;
}

/*********************************************************************
*
*       GUI_TOUCH_X_MeasureY()
*
* Function decription:
*   Called from GUI, if touch support is enabled.
*   Measures voltage of Y-axis.
*/
int  GUI_TOUCH_X_MeasureY(void) {
  return _TouchY;
}

#endif  // GUI_SUPPORT_TOUCH

