/* Copyright (c) 2012, DF9DQ
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list of conditions
 * and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions
 * and the following disclaimer in the documentation and/or other materials provided with the
 * distribution.
 * Neither the name of the author nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdint.h>

#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"

#include "bsp-hitex4300.h"

#include "lpclib.h"
#include "task-sys.h"
#include "task-gui.h"


/** \file
 *  \brief System management task for Hitex LPC4300 board.
 *
 *  The system management task is the heart of the control code. Its responsibility
 *  includes:
 *  - ....
 *
 *  \author NXP Semiconductors
 */



#define SYS_QUEUE_LENGTH                        (20)



typedef struct {
    uint8_t opcode;
    LPCLIB_Event event;
} SYS_Message;


/** Message opcodes for SYS task. */
enum {
    SYS_OPCODE_EVENT,
    SYS_OPCODE_BUTTON_ACTIVITY,             /**< Button activity (interrupt) */
};



struct {
    xQueueHandle queue;                     /**< Task message queue */
    int buttonFsm;                          /**< State machine to check button state */
    int buttonState;                        /**< Last state of buttons */
} sysContext;




/* System event handler. */
void SYS_eventHandler (LPCLIB_Event event)
{
    SYS_Message message;


    if (event.id != LPCLIB_EVENTID_APPLICATION) {
        return;
    }

    /* Handle things that can be done immediately without too much effort. */
    switch (event.opcode) {
    default:
        break;
    }

    /* Defer more complicated jobs to the SYS handler task. */
    message.opcode = SYS_OPCODE_EVENT;
    message.event = event;
    LPCLIB_OSALQueuePost(sysContext.queue, &message, LPCLIB_OSAL_QUEUE_BACK, LPCLIB_OSAL_TIMEOUT_NONE, NULL, LPCLIB_OSAL_CALL_DEFAULT);
}



/** Handle interrupt from buttons T1...T4 */
void GPIO0_IRQHandler (void)
{
    SYS_Message message;

    message.opcode = SYS_OPCODE_BUTTON_ACTIVITY;
    LPCLIB_OSALQueuePost(sysContext.queue, &message, LPCLIB_OSAL_QUEUE_BACK, LPCLIB_OSAL_TIMEOUT_NONE, NULL, LPCLIB_OSAL_CALL_ISR);

    GPIO_acknowledgeIndividualInterrupt(GPIO_INDIVIDUAL_INT_0);
}



/** Must be called to check button status. */
static void APP_handleButtons (void)
{
    uint16_t buttons;
    LPCLIB_Event event;
    int i;

    PCA9673_read(portExpander, &buttons);

    for (i = 0; i < 8; i++) {
        if (buttons ^ sysContext.buttonState ^ (0x100 << i)) {
            event.parameter = (buttons & (0x100 << i)) ? (void *)1 : (void *)0;
            switch (i) {
            case 0: event.channel = GUI_BUTTON_UP; break;
            case 1: event.channel = GUI_BUTTON_DOWN; break;
            case 2: event.channel = GUI_BUTTON_CENTER; break;
            case 3: event.channel = GUI_BUTTON_LEFT; break;
            case 4: event.channel = GUI_BUTTON_RIGHT; break;
            case 5: event.channel = GUI_BUTTON_SW3; break;
            case 6: event.channel = GUI_BUTTON_SW5; break;
            case 7: event.channel = GUI_BUTTON_SW6; break;
            }

            event.id = LPCLIB_EVENTID_APPLICATION;
            event.opcode = GUI_EVENT_BUTTON;
            GUI_sendEvent(event);
        }
    }

    sysContext.buttonState = buttons;
}



static const GPIO_Config configButtonInterrupt[] = {
    {.opcode = GPIO_OPCODE_CONFIGURE_INTERRUPT, /*.continued = DISABLE,*/
        {.interrupt = {
            .pin = BUTTON_IRQ_INPUT,
            .enable = ENABLE,
            .mode = GPIO_INT_FALLING_EDGE,
            .interruptLine = GPIO_INDIVIDUAL_INT_0, }}},
};



/* Submit a job for the system handler. */
void SYS_submitJob (LPCLIB_Event event)
{
    SYS_Message message;


    if (event.id != LPCLIB_EVENTID_APPLICATION) {
        return;
    }

    /* Handle things that can be done immediately without too much effort. */
    switch (event.opcode) {
    default:
        break;
    }

    /* Defer more complicated jobs to the SYS handler task. */
    message.opcode = SYS_OPCODE_EVENT;
    message.event = event;
    LPCLIB_OSALQueuePost(sysContext.queue, &message, LPCLIB_OSAL_QUEUE_BACK, LPCLIB_OSAL_TIMEOUT_NONE, NULL, LPCLIB_OSAL_CALL_DEFAULT);
}



void SYS_task (void *pArgs)
{
    (void) pArgs;
    SYS_Message message;


    /* Register events which we want to see. */
    GPIO_ioctl(&configButtonInterrupt[0]);


    while (1) {
        /* Is there a new message? */
        if (LPCLIB_OSALQueuePend(sysContext.queue,
                                 &message,
                                 LPCLIB_OSAL_TIMEOUT_INFINITE,
                                 NULL,
                                 LPCLIB_OSAL_CALL_DEFAULT) == LPCLIB_OSAL_ERROR_NONE) {
            switch (message.opcode) {
            case SYS_OPCODE_EVENT:
                switch(message.event.opcode) {
                }
                break;

            case SYS_OPCODE_BUTTON_ACTIVITY:
                APP_handleButtons();
                break;
            }
        }
    }
}


/** SYS module init function.
 *
 *  Must be called before the scheduler starts. Initializes the message queue.
 */
void SYS_initModule (void)
{
    /* Task communication */
    LPCLIB_OSALQueueCreate("", SYS_QUEUE_LENGTH, sizeof(SYS_Message), &sysContext.queue);
}

