
#ifndef __ISL29003_H
#define __ISL29003_H

#include <stdint.h>

#include "lpclib.h"


/** \defgroup ISL29003_Public_Types ISL29003 Types, enums, macros
 *  @{
 */


/** Opcodes to specify the configuration command in a call to \ref ISL29003_ioctl. */
typedef enum ISL29003_Opcode {
    ISL29003_OPCODE_POWER,                  /**< Config action: Power up/down */
    ISL29003_OPCODE_SET_INT_THRESHOLD,      /**< Config action: Set interrupt threshold */
} ISL29003_Opcode;


/** Descriptor to specify the configuration in a call to \ref ISL29003_ioctl. */
typedef struct ISL29003_Config {
    ISL29003_Opcode opcode;                 /**< Config action opcode */

    union {
        LPCLIB_Switch power;                /**< Config power state: ENABLE/DISABLE */
        uint32_t interruptThreshold;        /**< Config interrupt threshold */
    };
} ISL29003_Config;


/** Handle for an ISL29003 device. */
typedef struct ISL29003_Context *ISL29003_Handle;


/** Result type for lux value. */
typedef uint32_t ISL29003_LuxValue;



/** @} ISL29003 Types, enums, macros */



/** \defgroup ISL29003_Public_Functions ISL29003 API Functions
 *  @{
 */


/** Open the light sensor.
 *
 *  Enable access to a ISL29003 light sensor connected to an already open I2C bus.
 *
 *  \param[in] bus Handle of I2C bus to which the device is connected.
 *  \param[in] address I2C device address
 *  \param[out] pHandle Device handle
 *  \retval LPCLIB_SUCCESS Success. \ref pHandle contains a valid handle.
 */
LPCLIB_Result ISL29003_open (I2C_Handle bus, const uint8_t address, ISL29003_Handle *pHandle);


/** Configure the ISL29003 Light Sensor.
 *
 *  Pass a configuration command to the ISL29003. Configuration options
 *  include setting the gain or the interrupt threshold.
 *
 *  \param[in] handle Device handle as obtained by \ref ISL29003_open.
 *  \param[in] pConfig Pointer to a configuration descriptor of type
 *             \ref I2C_Config.
 */
LPCLIB_Result ISL29003_ioctl (ISL29003_Handle handle, const ISL29003_Config *pConfig);


/** Read lux value from light sensor.
 *
 *  Read raw conversion result from device, and convert to real lux value
 *  based on current settings.
 *
 *  \param[in] handle Device handle as obtained by \ref ISL29003_open.
 *  \param[in] pLux Pointer to a result.
 */
LPCLIB_Result ISL29003_read (ISL29003_Handle handle, ISL29003_LuxValue *pLux);


/** @} ISL29003 API Functions */

#endif
