
#ifndef __LM75_H
#define __LM75_H

#include <stdint.h>

#include "lpclib.h"


/** \defgroup LM75_Public_Types LM75 Types, enums, macros
 *  @{
 */


/** LM75 Register Names. */
enum LM75_RegisterNames {
    LM75_REG_TEMP = 0,
    LM75_REG_CONF = 1,
    LM75_REG_THYST = 2,
    LM75_REG_TOS = 3,

    __LM75_REG_LAST__ = LM75_REG_TOS,
};


/** Opcodes to specify the configuration command in a call to \ref LM75_ioctl. */
typedef enum LM75_Opcode {
    LM75_OPCODE_SET_THYST = LM75_REG_THYST,     /**< Config action: Thyst setting */
    LM75_OPCODE_SET_TOS = LM75_REG_TOS,         /**< Config action: Tos setting */
} LM75_Opcode;



/** Descriptor to specify the configuration in a call to \ref LM75_ioctl. */
typedef struct LM75_Config {
    LM75_Opcode opcode;                     /**< Config action opcode */
    LPCLIB_Switch continued;                /**< Set if further config struct's follow in an array */

    union {
        uint16_t temperature8_8;            /**< Temperature in C (format 8.8) */
    };
} LM75_Config;


/** Handle for an LM75 device. */
typedef struct LM75_Context *LM75_Handle;



/** @} LM75 Types, enums, macros */


/** \defgroup LM75_Public_Functions LM75 API Functions
 *  @{
 */


/** Open the device.
 *
 *  Enable access to an LM75 temperature sensor connected to an already open I2C bus.
 *
 *  \param[in] bus Handle of I2C bus to which the device is connected.
 *  \param[in] address I2C device address
 *  \param[out] pHandle Device handle
 *  \retval LPCLIB_SUCCESS Success. \ref pHandle contains a valid handle.
 */
LPCLIB_Result LM75_open (I2C_Handle bus, const uint8_t address, LM75_Handle *pHandle);


/** Configure the device.
 *
 *  Pass a configuration command to the LM75.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pConfig Pointer to a configuration descriptor
 */
LPCLIB_Result LM75_ioctl (LM75_Handle handle, const LM75_Config *pConfig);


/** Read current temperature.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pTemperature Pointer to result
 */
LPCLIB_Result LM75_read (LM75_Handle handle, uint16_t *pTemperature);


/** @} LM75 API Functions */

#endif
