
#ifndef __PCA9502_H
#define __PCA9502_H

#include <stdint.h>

#include "lpclib.h"

#if LPCLIB_I2C

/** \defgroup PCA9502_Public_Types PCA9502 Types, enums, macros
 *  @{
 */


/** Pin names of port expander. */
typedef enum PCA9502_Pin {
    PCA9502_PIN0 = 0,
    PCA9502_PIN1 = 1,
    PCA9502_PIN2 = 2,
    PCA9502_PIN3 = 3,
    PCA9502_PIN4 = 4,
    PCA9502_PIN5 = 5,
    PCA9502_PIN6 = 6,
    PCA9502_PIN7 = 7,
} PCA9502_Pin;



/** Opcodes to specify the configuration command in a call to \ref PCA9502_ioctl. */
typedef enum PCA9502_Opcode {
    PCA9502_OPCODE_RESET,                   /**< Config action: Reset device */
    PCA9502_OPCODE_SET_PIN_DIRECTION,       /**< Config action: Select input/output */
} PCA9502_Opcode;


/** Config option: Pin function */
typedef enum PCA9502_PinFunction {
    PCA9502_PINFUNC_INPUT,                  /**< Input */
    PCA9502_PINFUNC_OUTPUT,                 /**< Output */
} PCA9502_PinFunction;
struct PCA9502_ConfigPinFunc {
    PCA9502_Pin pin;                        /**< Pin number */
    PCA9502_PinFunction func;               /**< Function selector */
    LPCLIB_Switch enableInterrupt;          /**< Enable interrupt for input pin */
};




/** Descriptor to specify the configuration in a call to \ref PCA9502_ioctl. */
typedef struct PCA9502_Config {
    PCA9502_Opcode opcode;                  /**< Config action opcode */
    LPCLIB_Switch continued;                /**< Set if further config struct's follow in an array */

    union {
        struct PCA9502_ConfigPinFunc pinFunc;   /**< Pin function */
    };
} PCA9502_Config;


/** Handle for a PCA9502 device. */
typedef struct PCA9502_Context *PCA9502_Handle;



/** @} PCA9502 Types, enums, macros */


/** \defgroup PCA9502_Public_Functions PCA9502 API Functions
 *  @{
 */


/** Open the device.
 *
 *  Enable access to a PCA9502 port expander.
 *
 *  \param[in] bus Handle of I2C bus to which the device is connected.
 *  \param[in] address I2C device address
 *  \param[out] pHandle Device handle
 *  \retval LPCLIB_SUCCESS Success. \ref pHandle contains a valid handle.
 */
LPCLIB_Result PCA9502_open (I2C_Handle bus, const uint8_t address, PCA9502_Handle *pHandle);


/** Configure the device.
 *
 *  Pass a configuration command to the PCA9502.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pConfig Pointer to a configuration descriptor
 */
LPCLIB_Result PCA9502_ioctl (PCA9502_Handle handle, const PCA9502_Config *pConfig);


/** Read all input pins.
 *
 *  Read as one (half-)word the state of all input pins.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pPortState Pointer to result
 *  \retval LPCLIB_SUCCESS Ok
 *  \retval LPCLIB_.....
 */
LPCLIB_Result PCA9502_read (PCA9502_Handle handle, uint8_t *pPortState);


/** Return current state of a single port pin.
 *
 *  \param handle Device handle
 *  \param pin Pin number
 *  \return (32-bit) value with state (0/1) of the port pin.
 */
uint32_t PCA9502_readBit (PCA9502_Handle handle, PCA9502_Pin pin);


/** Write value (0/1) to a PCA9502 port pin.
 *
 *  \param[in] handle Port expander dvice handle
 *  \param[in] pin Pin selector
 *  \param[in] value Binary value to be written to port pin.
 *  \retval LPCLIB_SUCCESS Ok
 *  \retval LPCLIB_.....
 */
LPCLIB_Result PCA9502_writeBit (PCA9502_Handle handle, PCA9502_Pin pin, uint8_t value);


/** Write value to the 8-bit port through a mask.
 *
 *  \param[in] handle Port expander device handle
 *  \param[in] value 8-bit value to be written to port pins.
 *  \param[in] mask 8-bit value with a mask for parameter \a value. Only those
 *                  bits of \a value that have a 1 in the corresponding position
 *                  of \a mask take effect.
 *  \retval LPCLIB_SUCCESS Ok
 *  \retval LPCLIB_.....
 */
LPCLIB_Result PCA9502_write8WithMask (PCA9502_Handle handle, uint8_t value, uint8_t mask);



/** @} PCA9502 API Functions */

#endif  /* #if LPCLIB_I2C */

#endif  /* #ifndef __PCA9502_H */

