
#ifndef __PCA9532_H
#define __PCA9532_H

#include <stdint.h>

#include "lpclib.h"

#if LPCLIB_I2C

/** \defgroup PCA9532_Public_Types PCA9532 Types, enums, macros
 *  @{
 */


/** PCA9532 Register Names. */
enum PCA9532_RegisterNames {
    PCA9532_REG_INPUT0 = 0,
    PCA9532_REG_INPUT1 = 1,
    PCA9532_REG_PSC0 = 2,
    PCA9532_REG_PWM0 = 3,
    PCA9532_REG_PSC1 = 4,
    PCA9532_REG_PWM1 = 5,
    PCA9532_REG_LS0 = 6,
    PCA9532_REG_LS1 = 7,
    PCA9532_REG_LS2 = 8,
    PCA9532_REG_LS3 = 9,

    __PCA9532_REG_LAST__ = PCA9532_REG_LS3,
};


/** Pin names of port expander. */
typedef enum PCA9532_Pin {
    PCA9532_PIN0 = 0,
    PCA9532_PIN1 = 1,
    PCA9532_PIN2 = 2,
    PCA9532_PIN3 = 3,
    PCA9532_PIN4 = 4,
    PCA9532_PIN5 = 5,
    PCA9532_PIN6 = 6,
    PCA9532_PIN7 = 7,
    PCA9532_PIN8 = 8,
    PCA9532_PIN9 = 9,
    PCA9532_PIN10 = 10,
    PCA9532_PIN11 = 11,
    PCA9532_PIN12 = 12,
    PCA9532_PIN13 = 13,
    PCA9532_PIN14 = 14,
    PCA9532_PIN15 = 15,
} PCA9532_Pin;



/** Opcodes to specify the configuration command in a call to \ref PCA9532_ioctl. */
typedef enum PCA9532_Opcode {
    PCA9532_OPCODE_SET_PWM0 = PCA9532_REG_PSC0, /**< Config action: PWM0 setting */
    PCA9532_OPCODE_SET_PWM1 = PCA9532_REG_PSC1, /**< Config action: PWM1 setting */
    PCA9532_OPCODE_SET_PIN_FUNCTION,        /**< Config action: Select input/output/PWM0/PWM1 */
} PCA9532_Opcode;


/** Config option: PWM */
typedef enum PCA9532_PwmChannel {
    PCA9532_PWM0,                           /**< PWM channel 0 */
    PCA9532_PWM1,                           /**< PWM channel 1 */
} PCA9532_PwmChannel;
struct PCA9532_ConfigPwm {
    uint8_t period;                         /**< Period */
    uint8_t dutyCycle;                      /**< Duty cycle (128 = 50%) */
};


/** Config option: Pin function */
typedef enum PCA9532_PinFunction {
    PCA9532_PINFUNC_OFF,                    /**< OFF (high impedance) */
    PCA9532_PINFUNC_ON,                     /**< ON (driving low) */
    PCA9532_PINFUNC_PWM0,                   /**< Pin driven by PWM0 */
    PCA9532_PINFUNC_PWM1,                   /**< Pin driven by PWM1 */
} PCA9532_PinFunction;

struct PCA9532_ConfigPinFunc {
    uint8_t number;                         /**< Pin number */
    PCA9532_PinFunction func;               /**< Function selector */
};



/** Descriptor to specify the configuration in a call to \ref PCA9532_ioctl. */
typedef struct PCA9532_Config {
    PCA9532_Opcode opcode;                  /**< Config action opcode */
    LPCLIB_Switch continued;                /**< Set if further config struct's follow in an array */

    union {
        uint16_t direction;                 /**< Config pin direction (all pins!): ENABLE/DISABLE */
        struct PCA9532_ConfigPwm pwm;       /**< PWM configuration */
        struct PCA9532_ConfigPinFunc pinFunc;   /**< Pin function */
    };
} PCA9532_Config;


/** Handle for a PCA9532 device. */
typedef struct PCA9532_Context *PCA9532_Handle;



/** @} PCA9532 Types, enums, macros */


/** \defgroup PCA9532_Public_Functions PCA9532 API Functions
 *  @{
 */


/** Open the device.
 *
 *  Enable access to a PCA9532 port expander connected to an already open I2C bus.
 *
 *  \param[in] bus Handle of I2C bus to which the device is connected.
 *  \param[in] address I2C device address
 *  \param[out] pHandle Device handle
 *  \retval LPCLIB_SUCCESS Success. \ref pHandle contains a valid handle.
 */
LPCLIB_Result PCA9532_open (I2C_Handle bus, const uint8_t address, PCA9532_Handle *pHandle);


/** Configure the device.
 *
 *  Pass a configuration command to the PCA9532.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pConfig Pointer to a configuration descriptor
 */
LPCLIB_Result PCA9532_ioctl (PCA9532_Handle handle, const PCA9532_Config *pConfig);


/** Read all input pins.
 *
 *  Read as one (half-)word the state of all input pins.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pPortState Pointer to result
 */
LPCLIB_Result PCA9532_read (PCA9532_Handle handle, uint16_t *pPortState);


/** Return current state of a single port pin.
 *
 *  \param handle Device handle
 *  \param pin Pin number
 *  \return (32-bit) value with state (0/1) of the port pin.
 */
uint32_t PCA9532_readBit (PCA9532_Handle handle, PCA9532_Pin pin);


/** Write value (0/1) to a PCA9532 port pin.
 *
 *  \param[in] handle Port expander dvice handle
 *  \param[in] pin Pin selector
 *  \param[in] value Binary value to be written to port pin.
 */
void PCA9532_writeBit (PCA9532_Handle handle, PCA9532_Pin pin, uint8_t value);


/** Write value to the 16-bit port through a mask.
 *
 *  \param[in] handle Port expander device handle
 *  \param[in] value 16-bit value to be written to port pins.
 *  \param[in] mask 16-bit value with a mask for parameter \a value. Only those
 *                  bits of \a value that have a 1 in the corresponding position
 *                  of \a mask take effect.
 */
void PCA9532_write16WithMask (PCA9532_Handle handle, uint16_t value, uint16_t mask);



/** @} PCA9532 API Functions */

#endif  /* #if LPCLIB_I2C */

#endif  /* #ifndef __PCA9532_H */

