
#ifndef __PCA9673_H
#define __PCA9673_H

#include <stdint.h>

#include "lpclib.h"

#if LPCLIB_I2C

/** \defgroup PCA9673_Public_Types PCA9673 Types, enums, macros
 *  @{
 */


/** Pin names of port expander. */
typedef enum PCA9673_Pin {
    PCA9673_PIN00 = 0,
    PCA9673_PIN01 = 1,
    PCA9673_PIN02 = 2,
    PCA9673_PIN03 = 3,
    PCA9673_PIN04 = 4,
    PCA9673_PIN05 = 5,
    PCA9673_PIN06 = 6,
    PCA9673_PIN07 = 7,
    PCA9673_PIN10 = 8,
    PCA9673_PIN11 = 9,
    PCA9673_PIN12 = 10,
    PCA9673_PIN13 = 11,
    PCA9673_PIN14 = 12,
    PCA9673_PIN15 = 13,
    PCA9673_PIN16 = 14,
    PCA9673_PIN17 = 15,
} PCA9673_Pin;



/** Opcodes to specify the configuration command in a call to \ref PCA9673_ioctl. */
typedef enum PCA9673_Opcode {
    PCA9673_OPCODE_SET_PIN_DIRECTION,        /**< Config action: Select input/output */
} PCA9673_Opcode;


/** Config option: Pin function */
typedef enum PCA9673_PinFunction {
    PCA9673_PINFUNC_INPUT,                  /**< Input (high impedance) */
    PCA9673_PINFUNC_OUTPUT,                 /**< Allowed to be used as output */
} PCA9673_PinFunction;

struct PCA9673_ConfigPinFunc {
    PCA9673_Pin pin;                        /**< Pin number */
    PCA9673_PinFunction func;               /**< Function selector */
};



/** Descriptor to specify the configuration in a call to \ref PCA9673_ioctl. */
typedef struct PCA9673_Config {
    PCA9673_Opcode opcode;                  /**< Config action opcode */
    LPCLIB_Switch continued;                /**< Set if further config struct's follow in an array */

    union {
        uint16_t direction;                 /**< Config pin direction (all pins!): ENABLE/DISABLE */
        struct PCA9673_ConfigPinFunc pinFunc;   /**< Pin function */
    };
} PCA9673_Config;


/** Handle for a PCA9673 device. */
typedef struct PCA9673_Context *PCA9673_Handle;


/** @} PCA9673 Types, enums, macros */


/** \defgroup PCA9673_Public_Functions PCA9673 API Functions
 *  @{
 */


/** Open the device.
 *
 *  Enable access to a PCA9673 port expander
 *
 *  \param[in] bus Handle of I2C bus to which the device is connected.
 *  \param[in] address I2C device address
 *  \param[out] pHandle Device handle
 *  \retval LPCLIB_SUCCESS Success. \ref pHandle contains a valid handle.
 */
LPCLIB_Result PCA9673_open (I2C_Handle bus, const uint8_t address, PCA9673_Handle *pHandle);


/** Configure the device.
 *
 *  Pass a configuration command to the PCA9673.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pConfig Pointer to a configuration descriptor
 */
LPCLIB_Result PCA9673_ioctl (PCA9673_Handle handle, const PCA9673_Config *pConfig);


/** Read all input pins.
 *
 *  Read as one (half-)word the state of all input pins.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pPortState Pointer to result
 */
LPCLIB_Result PCA9673_read (PCA9673_Handle handle, uint16_t *pPortState);


/** Return current state of a single port pin.
 *
 *  \param handle Device handle
 *  \param pin Pin number
 *  \return (32-bit) value with state (0/1) of the port pin.
 */
uint32_t PCA9673_readBit (PCA9673_Handle handle, PCA9673_Pin pin);


/** Write value (0/1) to a PCA9673 port pin.
 *
 *  \param[in] handle Port expander dvice handle
 *  \param[in] pin Pin selector
 *  \param[in] value Binary value to be written to port pin.
 */
LPCLIB_Result PCA9673_writeBit (PCA9673_Handle handle, PCA9673_Pin pin, uint8_t value);


/** Write value to the 16-bit port through a mask.
 *
 *  \param[in] handle Port expander device handle
 *  \param[in] value 16-bit value to be written to port pins.
 *  \param[in] mask 16-bit value with a mask for parameter \a value. Only those
 *                  bits of \a value that have a 1 in the corresponding position
 *                  of \a mask take effect.
 */
LPCLIB_Result PCA9673_write16WithMask (PCA9673_Handle handle, uint16_t value, uint16_t mask);



/** @} PCA9673 API Functions */

#endif  /* #if LPCLIB_I2C */

#endif  /* #ifndef __PCA9673_H */

