
#ifndef __UDA1380_H
#define __UDA1380_H

#include <stdint.h>

#include "lpclib.h"


/** \defgroup UDA1380_Public_Types UDA1380 Types, enums, macros
 *  @{
 */


/** Opcodes to specify the configuration command in a call to \ref UDA1380_ioctl. */
typedef enum UDA1380_Opcode {
    UDA1380_OPCODE_CLOCKMODE,               /**< Config action: Clock options */
} UDA1380_Opcode;



/** Options for ADC activation. */
typedef enum UDA1380_AdcActivation {
    UDA1380_ADC_OFF = 0,                    /**< ADC disabled and powered down. */
    UDA1380_ADC_FROM_WSPLL = 0xC20,         /**< ADC active, uses WSPLL. */
    UDA1380_ADC_FROM_SYSCLK = 0xC00,        /**< ADC active, uses SYSCLK. */
} UDA1380_AdcActivation;

/** Options for DAC activation. */
typedef enum UDA1380_DacActivation {
    UDA1380_DAC_OFF = 0,                    /**< DAC disabled and powered down. */
    UDA1380_DAC_FROM_WSPLL = 0x310,         /**< DAC active, uses WSPLL. */
    UDA1380_DAC_FROM_SYSCLK = 0x300,        /**< DAC active, uses SYSCLK. */
} UDA1380_DacActivation;

/** Options for SYSCLK frequency. */
typedef enum UDA1380_SysclkFrequency {
    UDA1380_SYSCLK_256FS = 0,               /**< SYSCLK is 256 fs. */
    UDA1380_SYSCLK_384FS = 1,               /**< SYSCLK is 384 fs. */
    UDA1380_SYSCLK_512FS = 2,               /**< SYSCLK is 512 fs. */
    UDA1380_SYSCLK_768FS = 3,               /**< SYSCLK is 768 fs. */
    UDA1380_SYSCLK_DEFAULT = UDA1380_SYSCLK_256FS,
} UDA1380_SysclkFrequency;

/** Options for WSI (WSPLL) frequency. */
typedef enum UDA1380_WsiFrequency {
    UDA1380_WSI_6K25_12K5 = 0,              /**< WSI is 6.25...12.5 kHz. */
    UDA1380_WSI_12K5_25K = 1,               /**< WSI is 12.5...25 kHz. */
    UDA1380_WSI_25K_50K = 2,                /**< WSI is 25...50 kHz. */
    UDA1380_WSI_50K_100K = 3,               /**< WSI is 50...100 kHz. */
    UDA1380_WSI_DEFAULT = UDA1380_WSI_25K_50K,
} UDA1380_WsiFrequency;

/** Clock config. */
typedef uint16_t UDA1380_ClockConfigValue;



/** Descriptor to specify the configuration in a call to \ref LM75_ioctl. */
typedef struct UDA1380_Config {
    UDA1380_Opcode opcode;                  /**< Config action opcode */
    LPCLIB_Switch continued;                /**< Set if further config struct's follow in an array */

    union {
        UDA1380_ClockConfigValue clock;     /**< Clock configuration, global ADC/DAC power switch. */
    };
} UDA1380_Config;


/** Handle for a UDA1380 device. */
typedef struct UDA1380_Context *UDA1380_Handle;

/** @} UDA1380 Types, enums, macros */


/** \defgroup UDA1380_Public_Functions UDA1380 API Functions
 *  @{
 */


/** Open the device.
 *
 *  Enable access to a UDA1380 audio codec.
 *
 *  \param[in] bus Handle of I2C bus to which the device is connected.
 *  \param[in] address I2C device address
 *  \param[out] pHandle Device handle
 *  \retval LPCLIB_SUCCESS Success. \ref pHandle contains a valid handle.
 */
LPCLIB_Result UDA1380_open (I2C_Handle bus, const uint8_t address, UDA1380_Handle *pHandle);


/** Configure the device.
 *
 *  Pass a configuration command to the UDA1380.
 *
 *  \param[in] handle Device handle.
 *  \param[in] pConfig Pointer to a configuration descriptor
 */
LPCLIB_Result UDA1380_ioctl (UDA1380_Handle handle, const UDA1380_Config *pConfig);


/** Helper function to create a clock configuration from the settings.
 *
 *  \param[in] wspll Enable/disable use of WSPLL
 *  \param[in] adcMode Select ADC activation and clock source
 *  \param[in] dacMode Select DAC activation and clock source
 *  \return All settings combined in one word
 */
static UDA1380_ClockConfigValue UDA1380_makeConfigClock (
        UDA1380_AdcActivation adc,
        UDA1380_DacActivation dac,
        UDA1380_SysclkFrequency sysclkFrequency,
        UDA1380_WsiFrequency wsiFrequency);


/** @} UDA1380 API Functions */



__FORCEINLINE(UDA1380_ClockConfigValue UDA1380_makeConfigClock (
        UDA1380_AdcActivation adc,
        UDA1380_DacActivation dac,
        UDA1380_SysclkFrequency sysclkFrequency,
        UDA1380_WsiFrequency wsiFrequency))
{
    return (UDA1380_ClockConfigValue)(
        (wsiFrequency << 0)     |
        (sysclkFrequency << 2)  |
        (dac << 9)              |
        (adc << 11)
        );
}

#endif


