
#include <stdint.h>
#include <string.h>

#include "lpclib.h"

#include "at45db081d.h"


/** Local device context. */
typedef struct AT45_Context {
    SSP_Handle bus;                         /**< Bus to which device is connected. */
    int16_t pageSize;                       /**< 256 Byte vs. 264 Byte page mode */
    const SSP_DeviceSelect *pDeviceSelect;  /**< SSEL demuxing. */
    int32_t filepos;                        /**< Position of read/write pointer */
    int32_t filesize;                       /**< Total number of bytes in flash */
} AT45_Context;


static AT45_Context globalContext; //TODO: Must allocate memory dynamically!!




const SSP_Config at45Device[] = {
    {.opcode = SSP_OPCODE_SET_FORMAT, .continued = ENABLE,
        {.format = {.bits = SSP_BITS_8,
                    .clockFormat = SSP_CLOCK_MODE3,
                    .frameFormat = SSP_FRAMEFORMAT_SPI, }}},
    {.opcode = SSP_OPCODE_SET_BITRATE, .continued = DISABLE,
//        {.bitrate = 24000000ul, }},
        {.bitrate = 12000000ul, }},
};



/* Open access to flash device */
LPCLIB_Result AT45_open (SSP_Handle bus, const SSP_DeviceSelect *pDeviceSelect, AT45_Handle *pHandle)
{
    AT45_Context *pContext = &globalContext; //TODO


    pContext->bus = bus;
    pContext->pDeviceSelect = pDeviceSelect;
    pContext->filepos = 0;
    pContext->pageSize = 264;    //TODO derive from status register
    pContext->filesize = 0x100000ul; //TODO

    *pHandle = pContext;

    return LPCLIB_SUCCESS;
}



/* Close access to flash device */
LPCLIB_Result AT45_close (AT45_Handle *pHandle)
{
    if (*pHandle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    *pHandle = LPCLIB_INVALID_HANDLE;

    return LPCLIB_SUCCESS;
}



/* Set file pointer */
LPCLIB_Result AT45_lseek (AT45_Handle handle, int32_t offset, LPCLIB_FilePos whence)
{
    if (handle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    switch (whence) {
    case LPCLIB_SEEK_SET:
        handle->filepos = offset;
        break;
    case LPCLIB_SEEK_CUR:
        handle->filepos += offset;
        if (handle->filepos >= handle->filesize) {
            handle->filepos = handle->filesize;
        }
        break;
    case LPCLIB_SEEK_END:
        handle->filepos = handle->filesize - offset;
        if (handle->filepos < 0) {
            handle->filepos = 0;
        }
        break;
    }

    return LPCLIB_SUCCESS;
}



/* Read from current position onwards */
LPCLIB_Result AT45_read (AT45_Handle handle, void *rxBuffer, size_t numBytes)
{
    SSP_Job job;
    SSP_JobPhase phase[2];
    uint8_t buf[8];
    uint32_t pageNum;


    if (handle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    pageNum = handle->filepos / handle->pageSize;

    buf[0] = 0x0B;                                      /* command: continuous array read */
    if (handle->pageSize == 264) {
        buf[1] = (pageNum >> 7) & 0x1F;
        buf[2] = ((pageNum << 1) & 0xFE) | (((handle->filepos % handle->pageSize) >> 8) & 1);
    }
    else {
        buf[1] = (pageNum >> 8) & 0x0F;
        buf[2] = (pageNum     ) & 0xFF;
    }
    buf[3] = handle->filepos % handle->pageSize;
    buf[4] = 0;

    job.pDeviceSelect = handle->pDeviceSelect;
    job.pConfig = &at45Device[0];
    job.firstPhase = SSP_makePhase(&phase[0], &buf[0], NULL, 5);
    phase[0].next = SSP_makePhase(&phase[1], NULL, rxBuffer, numBytes);

    if (SSP_submitJob(handle->bus, &job) != LPCLIB_SUCCESS) {
        return LPCLIB_ERROR;
    }

    handle->filepos = (handle->filepos + numBytes) % handle->filesize;

    return LPCLIB_SUCCESS;
}



/* Write to current position onwards */
LPCLIB_Result AT45_write (AT45_Handle handle, const void *txBuffer, size_t numBytes)
{
    SSP_Job job;
    SSP_JobPhase phase[2];
    uint8_t buf[8];
    uint32_t temp;
    uint32_t pageNum;


    if (handle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    pageNum = handle->filepos / handle->pageSize;

    while (numBytes > 0) {
        /* If the next page is not a full 256 byte page (a partial page at the beginning
         * or the end of the write block), we copy the existing content to a buffer first.
         */
        if ((handle->filepos % 256) || (numBytes < 256)) {
            buf[0] = 0x53;                                  /* command: main memory to buffer 1 */
            if (handle->pageSize == 264) {
                buf[1] = (pageNum >> 7) & 0x1F;
                buf[2] = (pageNum << 1) & 0xFE;
            }
            else {
                buf[1] = (pageNum >> 8) & 0x0F;
                buf[2] = (pageNum     ) & 0xFF;
            }
            buf[3] = 0;

            job.pDeviceSelect = handle->pDeviceSelect;
            job.pConfig = &at45Device[0];
            job.firstPhase = SSP_makePhase(&phase[0], &buf[0], NULL, 4);

            if (SSP_submitJob(handle->bus, &job) != LPCLIB_SUCCESS) {
                return LPCLIB_ERROR;
            }
        }

        /* Fill buffer up to the next page boundary (or buffer end) */
        temp = 256 - (numBytes % 256);

        buf[0] = 0x84;                                      /* command: write to buffer 1 */
        buf[1] = 0;
        buf[2] = 0;
        buf[3] = handle->filepos % 256;

        job.pDeviceSelect = handle->pDeviceSelect;
        job.pConfig = &at45Device[0];
        job.firstPhase = SSP_makePhase(&phase[0], &buf[0], NULL, 4);
        phase[0].next = SSP_makePhase(&phase[1], txBuffer, NULL, temp);

        if (SSP_submitJob(handle->bus, &job) != LPCLIB_SUCCESS) {
            return LPCLIB_ERROR;
        }

        /* Copy buffer to main memory */
        buf[0] = 0x83;                                  /* command: buffer 1 to main memory */
        if (handle->pageSize == 264) {
            buf[1] = (pageNum >> 7) & 0x1F;
            buf[2] = (pageNum << 1) & 0xFE;
        }
        else {
            buf[1] = (pageNum >> 8) & 0x0F;
            buf[2] = (pageNum     ) & 0xFF;
        }
        buf[3] = 0;

        job.pDeviceSelect = handle->pDeviceSelect;
        job.pConfig = &at45Device[0];
        job.firstPhase = SSP_makePhase(&phase[0], &buf[0], NULL, 4);

        if (SSP_submitJob(handle->bus, &job) != LPCLIB_SUCCESS) {
            return LPCLIB_ERROR;
        }

        numBytes -= temp;
        handle->filepos = (handle->filepos + temp) % handle->filesize;
    }

    return LPCLIB_SUCCESS;
}


