
#include <stdint.h>
#include <string.h>

#include "lpclib.h"

#if LPCLIB_I2C

#include "i2cdev.h"
#include "pca9502.h"


/** PCA9502 Register Names. */
enum PCA9502_RegisterNames {
    PCA9502_REG_IODIR = 10,
    PCA9502_REG_IOSTATE = 11,
    PCA9502_REG_IOINTENA = 12,
    PCA9502_REG_IOCONTROL = 14,

    __PCA9502_REG_LAST__ = PCA9502_REG_IOCONTROL,
};


LPCLIB_DefineRegBit(PCA9502_IOCONTROL_SRESET,   3,  1);


/** Local device context. */
typedef struct PCA9502_Context {
    I2C_Handle bus;                         /**< Bus to which device is connected. */
    uint8_t slaveAddress;                   /**< Slave address on I2C bus. */
} PCA9502_Context;


static PCA9502_Context globalContext; //TODO: Must allocate memory dynamically!!


/** Read a single register.
 *
 *  \param[in] handle Handle of port expander.
 *  \param[in] theReg Register from which to read.
 *  \param[out] pReg Points to buffer that takes register content.
 *  \retval LPCLIB_SUCCESS Ok!
 *  \retval .... Error
 */
static LPCLIB_Result PCA9502_readReg (PCA9502_Handle handle,
                                      enum PCA9502_RegisterNames theReg,
                                      uint8_t *pReg)
{
    uint8_t buf;

    buf = theReg << 3;
    return I2CDEV_writeAndRead(handle->bus, handle->slaveAddress, 1, &buf, 1, pReg);
}



/** Write one register.
 *
 *  \param[in] handle Handle of port expander.
 *  \param[in] theReg Register to be written to.
 *  \param[out] pRegs Points to buffer that takes register content.
 *  \retval LPCLIB_SUCCESS Ok!
 *  \retval .... Error
 */
static LPCLIB_Result PCA9502_writeReg (PCA9502_Handle handle,
                                       enum PCA9502_RegisterNames theReg,
                                       const uint8_t *pReg)
{
    uint8_t buf[2];

    buf[0] = theReg << 3;
    buf[1] = *pReg;
    return I2CDEV_write(handle->bus, handle->slaveAddress, 2, &buf);
}



LPCLIB_Result PCA9502_open (I2C_Handle bus, const uint8_t address, PCA9502_Handle *pHandle)
{
    PCA9502_Context *pContext = &globalContext; //TODO!


    pContext->bus = bus;
    pContext->slaveAddress = address;

    *pHandle = pContext;

    return LPCLIB_SUCCESS;
}



LPCLIB_Result PCA9502_ioctl (PCA9502_Handle handle, const PCA9502_Config *pConfig)
{
    LPCLIB_Result result = LPCLIB_SUCCESS;
    uint8_t reg;


    if (handle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    while (1) {
        switch (pConfig->opcode) {
        case PCA9502_OPCODE_RESET:
            reg = PCA9502_IOCONTROL_SRESET_Msk;
            result = PCA9502_writeReg(handle, PCA9502_REG_IOCONTROL, &reg);
            break;

        case PCA9502_OPCODE_SET_PIN_DIRECTION:
            result = PCA9502_readReg(handle, PCA9502_REG_IODIR, &reg);
            if (result == LPCLIB_SUCCESS) {
                if (pConfig->pinFunc.func == PCA9502_PINFUNC_OUTPUT) {
                    reg |= (1u << pConfig->pinFunc.pin);
                }
                else {
                    reg &= ~(1u << pConfig->pinFunc.pin);
                }

                result = PCA9502_writeReg(handle, PCA9502_REG_IODIR, &reg);
                if (result == LPCLIB_SUCCESS) {
                    if (pConfig->pinFunc.func == PCA9502_PINFUNC_INPUT) {
                        result = PCA9502_readReg(handle, PCA9502_REG_IOINTENA, &reg);
                        if (result == LPCLIB_SUCCESS) {
                            if (pConfig->pinFunc.enableInterrupt) {
                                reg |= (1u << pConfig->pinFunc.pin);
                            }
                            else {
                                reg &= ~(1u << pConfig->pinFunc.pin);
                            }
    
                            result = PCA9502_writeReg(handle, PCA9502_REG_IOINTENA, &reg);
                        }
                    }
                }
            }
        }

        if (result != LPCLIB_SUCCESS) {
            break;
        }

        if (!pConfig->continued) {
            break;
        }

        ++pConfig;                                      /* More config's follow */
    }

    return result;
}



/* Read all input pins. */
LPCLIB_Result PCA9502_read (PCA9502_Handle handle, uint8_t *pPortState)
{
    uint8_t reg;
    LPCLIB_Result result;


    if (handle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    result = PCA9502_readReg(handle, PCA9502_REG_IOSTATE, &reg);

    *pPortState = reg;

    return result;
}


/* Write value to the 8-bit port through a mask. */
LPCLIB_Result PCA9502_write8WithMask (PCA9502_Handle handle, uint8_t value, uint8_t mask)
{
    uint8_t reg;
    LPCLIB_Result result;

    result = PCA9502_readReg(handle, PCA9502_REG_IOSTATE, &reg);
    if (result == LPCLIB_SUCCESS) {
        reg = (reg & ~mask) | (value & mask);
        result = PCA9502_writeReg(handle, PCA9502_REG_IOSTATE, &reg);
    }

    return result;
}


/* Write value (0/1) to a port pin. */
LPCLIB_Result PCA9502_writeBit (PCA9502_Handle handle, PCA9502_Pin pin, uint8_t value)
{
    uint8_t reg;
    LPCLIB_Result result;


    result = PCA9502_read(handle, &reg);
    if (result == LPCLIB_SUCCESS) {
        if (value) {
            reg |= (1u << pin);
        }
        else {
            reg &= ~(1u << pin);
        }
        result = PCA9502_writeReg(handle, PCA9502_REG_IOSTATE, &reg);
    }
    
    return result;
}


/* Return current state of a single port pin. */
uint32_t PCA9502_readBit (PCA9502_Handle handle, PCA9502_Pin pin)
{
    uint8_t port;

    if (PCA9502_read(handle, &port) == LPCLIB_SUCCESS) {
        return (port >> pin) & 1;                       /* Extract requested pin */
    }

    return 0;                                           /* Error. Return something anyway */
}


#endif  /* #if LPCLIB_I2C */
