
#include <stdint.h>
#include <string.h>

#include "lpclib.h"

#if LPCLIB_I2C

#include "i2cdev.h"
#include "pca9532.h"


/** Local device context. */
typedef struct PCA9532_Context {
    I2C_Handle bus;                         /**< Bus to which device is connected. */
    uint8_t slaveAddress;                   /**< Slave address on I2C bus. */
} PCA9532_Context;


static PCA9532_Context globalContext; //TODO: Must allocate memory dynamically!!


/** Read a single register.
 *
 *  \param[in] handle Handle of port expander.
 *  \param[in] theReg Register from which to read.
 *  \param[out] pReg Points to buffer that takes register content.
 *  \retval LPCLIB_SUCCESS Ok!
 *  \retval .... Error
 */
static LPCLIB_Result PCA9532_readReg (PCA9532_Handle handle,
                                      enum PCA9532_RegisterNames theReg,
                                      uint8_t *pReg)
{
    uint8_t buf;

    buf = (1u << 4) + theReg;
    return I2CDEV_writeAndRead(handle->bus, handle->slaveAddress, 1, &buf, 1, pReg);
}



/** Write one register.
 *
 *  \param[in] handle Handle of port expander.
 *  \param[in] theReg Register to be written to.
 *  \param[out] pRegs Points to buffer that takes register content.
 *  \retval LPCLIB_SUCCESS Ok!
 *  \retval .... Error
 */
static LPCLIB_Result PCA9532_writeReg (PCA9532_Handle handle,
                                       enum PCA9532_RegisterNames theReg,
                                       const uint8_t *pReg)
{
    uint8_t buf[2];

    buf[0] = (1u << 4) + theReg;
    buf[1] = *pReg;
    return I2CDEV_write(handle->bus, handle->slaveAddress, 2, &buf);
}



LPCLIB_Result PCA9532_open (I2C_Handle bus, const uint8_t address, PCA9532_Handle *pHandle)
{
    PCA9532_Context *pContext = &globalContext; //TODO!


    pContext->bus = bus;
    pContext->slaveAddress = address;

    *pHandle = pContext;

    return LPCLIB_SUCCESS;
}



LPCLIB_Result PCA9532_ioctl (PCA9532_Handle handle, const PCA9532_Config *pConfig)
{
    enum PCA9532_RegisterNames regNum;
    LPCLIB_Result result = LPCLIB_SUCCESS;
    uint8_t reg;


    if (handle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    while (1) {
        switch (pConfig->opcode) {
        case PCA9532_OPCODE_SET_PWM0:
        case PCA9532_OPCODE_SET_PWM1:
            regNum = (enum PCA9532_RegisterNames)(pConfig->opcode);     /* The opcode is conveniently encoded! */
            result = PCA9532_writeReg(handle, regNum, &pConfig->pwm.period);
            if (result == LPCLIB_SUCCESS) {
                result = PCA9532_writeReg(handle,
                                          (enum PCA9532_RegisterNames)(regNum + 1),
                                          &pConfig->pwm.dutyCycle);  //TODO write both regs in one call!
            }
            break;
    
        case PCA9532_OPCODE_SET_PIN_FUNCTION:
            if (pConfig->pinFunc.number < 4) {
                regNum = PCA9532_REG_LS0;
            }
            else if (pConfig->pinFunc.number < 8) {
                regNum = PCA9532_REG_LS1;
            }
            else if (pConfig->pinFunc.number < 12) {
                regNum = PCA9532_REG_LS2;
            }
            else {
                regNum = PCA9532_REG_LS3;
            }
            result = PCA9532_readReg(handle, regNum, &reg);
            if (result == LPCLIB_SUCCESS) {
                reg &= ~(3u << (2 * (pConfig->pinFunc.number % 4)));
                reg |= (pConfig->pinFunc.func << (2 * (pConfig->pinFunc.number % 4)));
                result = PCA9532_writeReg(handle, regNum, &reg);
            }
            break;
        }

        if (!pConfig->continued) {
            break;
        }

        ++pConfig;                                      /* More config's follow */
    }

    return result;
}



LPCLIB_Result PCA9532_read (PCA9532_Handle handle, uint16_t *pPortState)
{
    uint8_t buf;
    uint8_t regs[2];
    LPCLIB_Result result;


    if (handle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    buf = PCA9532_REG_INPUT0;
    result = I2CDEV_writeAndRead(handle->bus, handle->slaveAddress, 1, &buf, 1, &regs[0]);
    if (result == LPCLIB_SUCCESS) {
        buf = PCA9532_REG_INPUT1;
        result = I2CDEV_writeAndRead(handle->bus, handle->slaveAddress, 1, &buf, 1, &regs[1]);
    }

    *pPortState = (regs[1] << 8) | regs[0];

    return result;
}


/* Write value to the 16-bit port through a mask. */
void PCA9532_write16WithMask (PCA9532_Handle handle, uint16_t value, uint16_t mask)
{
    int i;
    PCA9532_Config config;

    for (i = 0; i < 16; i++) {
        if (mask & (1u << i)) {
            config.opcode = PCA9532_OPCODE_SET_PIN_FUNCTION;
            config.continued = DISABLE;
            config.pinFunc.number = i;
            config.pinFunc.func = value & (1u << i) ?
                    PCA9532_PINFUNC_ON :
                    PCA9532_PINFUNC_OFF;
            PCA9532_ioctl(handle, &config);
        }
    }
}


/* Write value (0/1) to a PCA9532 port pin. */
void PCA9532_writeBit (PCA9532_Handle handle, PCA9532_Pin pin, uint8_t value)
{
    PCA9532_Config config;


    if (pin < 16) {
        config.opcode = PCA9532_OPCODE_SET_PIN_FUNCTION;
        config.continued = DISABLE;
        config.pinFunc.number = pin;
        config.pinFunc.func = value ? PCA9532_PINFUNC_ON : PCA9532_PINFUNC_OFF;
        PCA9532_ioctl(handle, &config);
    }
}


/* Return current state of a single port pin. */
uint32_t PCA9532_readBit (PCA9532_Handle handle, PCA9532_Pin pin)
{
    uint16_t port;

    if (PCA9532_read(handle, &port) == LPCLIB_SUCCESS) {
        return (port >> pin) & 1;                       /* Extract requested pin */
    }

    return 0;                                           /* Error. Return something anyway */
}

#endif  /* #if LPCLIB_I2C */
