
#include <stdint.h>
#include <string.h>

#include "lpclib.h"

#include "pcf2129-spi.h"



/** PCF2129 Register Names. */
enum PCF2129_RegisterNames {
    PCF2129_REG_CONTROL_1 = 0x00,
    PCF2129_REG_CONTROL_2 = 0x01,
    PCF2129_REG_CONTROL_3 = 0x02,
    PCF2129_REG_SECONDS = 0x03,
    PCF2129_REG_MINUTES = 0x04,
    PCF2129_REG_HOURS = 0x05,
    PCF2129_REG_DAYS = 0x06,
    PCF2129_REG_WEEKDAYS = 0x07,
    PCF2129_REG_MONTHS = 0x08,
    PCF2129_REG_YEARS = 0x09,
    PCF2129_REG_SECOND_ALARM = 0x0A,
    PCF2129_REG_MINUTE_ALARM = 0x0B,
    PCF2129_REG_HOUR_ALARM = 0x0C,
    PCF2129_REG_DAY_ALARM = 0x0D,
    PCF2129_REG_WEEKDAY_ALARM = 0x0E,
    PCF2129_REG_CLKOUT_CTL = 0x0F,
    PCF2129_REG_WATCHDOG_TIM_CTL = 0x10,
    PCF2129_REG_WATCHDOG_TIM_VAL = 0x11,
    PCF2129_REG_TIMESTP_CTL = 0x12,
    PCF2129_REG_SEC_TIMESTP = 0x13,
    PCF2129_REG_MIN_TIMESTP = 0x14,
    PCF2129_REG_HOUR_TIMESTP = 0x15,
    PCF2129_REG_DAY_TIMESTP = 0x16,
    PCF2129_REG_MON_TIMESTP = 0x17,
    PCF2129_REG_YEAR_TIMESTP = 0x18,
    PCF2129_REG_AGING_OFFSET = 0x19,

    __PCF2129_REG_LAST__ = PCF2129_REG_AGING_OFFSET,
};


/** Local device context. */
typedef struct PCF2129_Context {
    SSP_Handle bus;                         /**< Bus to which device is connected. */
    const SSP_DeviceSelect *pDeviceSelect;  /**< SSEL demuxing. */
} PCF2129_Context;


static PCF2129_Context globalContext; //TODO: Must allocate memory dynamically!!




static const SSP_Config pcf2129Device[] = {
    {.opcode = SSP_OPCODE_SET_FORMAT, .continued = ENABLE,
        {.format = {
            .bits = SSP_BITS_8,
            .clockFormat = SSP_CLOCK_CPHA0_CPOL0,
            .frameFormat = SSP_FRAMEFORMAT_SPI, }}},

    {.opcode = SSP_OPCODE_SET_BITRATE, .continued = DISABLE,
        {.bitrate = 1000000ul, }},
};





/** Read one or more registers.
 */
static LPCLIB_Result PCF2129_readReg (PCF2129_Handle handle,
                                      enum PCF2129_RegisterNames firstReg,
                                      enum PCF2129_RegisterNames lastReg,
                                      uint8_t *pReg)
{
    uint8_t command;
    SSP_Job job;
    SSP_JobPhase phase[2];


    command = 0xA0 + firstReg;
    job.firstPhase = SSP_makePhase(&phase[0], &command, NULL, 1);
    phase[0].next = SSP_makePhase(&phase[1], NULL, pReg, 1 + (lastReg - firstReg));

    job.pDeviceSelect = handle->pDeviceSelect;
    job.pConfig = &pcf2129Device[0];

    if (SSP_submitJob(handle->bus, &job) != LPCLIB_SUCCESS) {
        return LPCLIB_ERROR;
    }

    return LPCLIB_SUCCESS;
}



/** Write one or more registers.
 */
static LPCLIB_Result PCF2129_writeReg (PCF2129_Handle handle,
                                       enum PCF2129_RegisterNames firstReg,
                                       enum PCF2129_RegisterNames lastReg,
                                       uint8_t *pReg)
{
    uint8_t command;
    SSP_Job job;
    SSP_JobPhase phase[2];


    command = 0x20 + firstReg;
    job.firstPhase = SSP_makePhase(&phase[0], &command, NULL, 1);
    phase[0].next = SSP_makePhase(&phase[1], pReg, NULL, 1 + (lastReg - firstReg));

    job.pDeviceSelect = handle->pDeviceSelect;
    job.pConfig = &pcf2129Device[0];

    if (SSP_submitJob(handle->bus, &job) != LPCLIB_SUCCESS) {
        return LPCLIB_ERROR;
    }

    return LPCLIB_SUCCESS;
}



LPCLIB_Result PCF2129_open (SSP_Handle bus, const SSP_DeviceSelect *pDeviceSelect, PCF2129_Handle *pHandle)
{
    PCF2129_Context *pContext = &globalContext; //TODO


    pContext->bus = bus;
    pContext->pDeviceSelect = pDeviceSelect;

    *pHandle = pContext;

    return LPCLIB_SUCCESS;
}



/** Read current time and date.
 */
LPCLIB_Result PCF2129_readTimeDateCtime (PCF2129_Handle handle, PCF2129_TimeDateCtime *timedate)
{
    LPCLIB_Result  result;
    uint8_t regs[7];        /* Takes RTC registers 03h...09h */


    /* Read all time/date registers in one transaction (see data sheet!) */
    result = PCF2129_readReg (handle, PCF2129_REG_SECONDS, PCF2129_REG_YEARS, &regs[0]);
    if (result == LPCLIB_SUCCESS) {
        /* Convert from BCD into format used by LPC17xx RTC (registers CTIME0, CTIME1) */
        timedate->ctime0 =
                /* Seconds */
                ((uint32_t)(((regs[0] >> 0) & 0x0F) + 10 * ((regs[0] >> 4) & 0x07)) << 0) |
                /* Minutes */
                ((uint32_t)(((regs[1] >> 0) & 0x0F) + 10 * ((regs[1] >> 4) & 0x07)) << 8UL) |
                /* Hours */
                ((uint32_t)(((regs[2] >> 0) & 0x0F) + 10 * ((regs[2] >> 4) & 0x03)) << 16UL) |
                /* Day of week */
                (((uint32_t)regs[4] & 0x07) << 24UL);
        timedate->ctime1 =
                /* Day of Month */
                ((uint32_t)(((regs[3] >> 0) & 0x0F) + 10 * ((regs[3] >> 4) & 0x03)) << 0) |
                /* Month */
                ((uint32_t)(((regs[5] >> 0) & 0x0F) + 10 * ((regs[5] >> 4) & 0x01)) << 8UL) |
                /* Year */
                (((uint32_t)(((regs[6] >> 0) & 0x0F) + 10 * ((regs[6] >> 4) & 0x0F)) + 2000UL) << 16UL);
    }

    return result;
}


/** Read current time and date.
 */
LPCLIB_Result PCF2129_readTimeDate (PCF2129_Handle handle, PCF2129_TimeDate *timedate)
{
    /* Read all time/date registers in one transaction (see data sheet!) */
    return PCF2129_readReg (handle, PCF2129_REG_SECONDS, PCF2129_REG_YEARS, (uint8_t *)timedate);
}



/** Let the RTC run.
 */
LPCLIB_Result PCF2129_run (PCF2129_Handle handle)
{
    uint8_t control;

    control = 0;
    return PCF2129_writeReg(handle, PCF2129_REG_CONTROL_1, PCF2129_REG_CONTROL_1, &control);
}


#if 0
/** Set time and date. */
bool pcf2127_write_time_date (struct pcf2127_time_date *timedate)
{
    bool result = false;
    uint8_t regs[7];        /* Takes RTC registers 03h...09h */
    uint32_t value;


    /* STOP the RTC */
    regs[0] = (1 << 5);
    if (pcf2127_write_reg (0, 1, &regs[0]))
    {
        /* Convert from LPC17xx format into BCD format used by PCF2127 */
        value = (timedate->ctime0 >> 0) & 0x3F;
        regs[0] = 16 * (value / 10) + (value % 10);
        value = (timedate->ctime0 >> 8UL) & 0x3F;
        regs[1] = 16 * (value / 10) + (value % 10);
        value = (timedate->ctime0 >> 16UL) & 0x1F;
        regs[2] = 16 * (value / 10) + (value % 10);
        value = (timedate->ctime0 >> 24UL) & 0x07;
        regs[4] = value;
        value = (timedate->ctime1 >> 0) & 0x1F;
        regs[3] = 16 * (value / 10) + (value % 10);
        value = (timedate->ctime1 >> 8UL) & 0x0F;
        regs[5] = 16 * (value / 10) + (value % 10);
        value = ((timedate->ctime1 >> 16UL) & 0xFFF) - 2000;
        regs[6] = 16 * (value / 10) + (value % 10);

        if (pcf2127_write_reg (3, 7, &regs[0]))
        {
            result = true;
        }
    }

    /* START the RTC */
    regs[0] = (0 << 5);
    if (!pcf2127_write_reg (0, 1, &regs[0]))
    {
        result = false;
    }

    return result;
}
#endif
