/* $Id::                                                               $
 *
 * Copyright (c) 2011, NXP Semiconductors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file
 *  \brief CLKPWR driver interface.
 *  This file defines all interface objects needed to use the CLKPWR driver.
 *
 *  \author NXP Semiconductors
 */


#ifndef __LPC43xx_CLKPWR_H__
#define __LPC43xx_CLKPWR_H__

/** \defgroup CLKPWR
 *  \ingroup API
 *  @{
 */

#include "lpc43xx_libconfig.h"

#include "lpclib_types.h"


/** \defgroup CLKPWR_Public_Types CLKPWR Types, enums, macros
 *  @{
 */

typedef enum CLKPWR_Source {
    CLKPWR_SOURCE_RTC = 0,
    CLKPWR_SOURCE_IRC = 1,
    CLKPWR_SOURCE_ENET_RX = 2,
    CLKPWR_SOURCE_ENET_TX = 3,
    CLKPWR_SOURCE_GP_CLKIN = 4,
    CLKPWR_SOURCE_XTAL = 6,
    CLKPWR_SOURCE_PLL0_AUDIO = 8,
    CLKPWR_SOURCE_PLL1 = 9,
    CLKPWR_SOURCE_IDIVA = 12,
    CLKPWR_SOURCE_IDIVB = 13,
    CLKPWR_SOURCE_IDIVC = 14,
    CLKPWR_SOURCE_IDIVD = 15,
    CLKPWR_SOURCE_IDIVE = 16,
} CLKPWR_Source;


/** Base clocks. */
typedef enum CLKPWR_BaseClock {
    /* 30:     CCU number (0=CCU1, 1=CCU2)
     * 29-26:  Bit in CCUx_BASE_STAT
     * 25-16:  Register offset in CCUx
     * 15-11:  OUTCLK register number
     * 10:     Has base clock entry in CCUx
     */
    CLKPWR_BASECLOCK_APB3   = (0 << 30) | (0 << 26) | ( 64 << 16) | (10 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_APB1   = (0 << 30) | (1 << 26) | (128 << 16) | ( 9 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_SPIFI  = (0 << 30) | (2 << 26) | (192 << 16) | ( 5 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_M3     = (0 << 30) | (3 << 26) | (256 << 16) | ( 4 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_LCD    = (0 << 30) | (3 << 26)               | (11 << 11) | (0 << 10),
    CLKPWR_BASECLOCK_USB0   = (0 << 30) | (7 << 26) | (512 << 16) | ( 1 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_USB1   = (0 << 30) | (8 << 26) | (576 << 16) | ( 3 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_UART3  = (1 << 30) | (1 << 26) | (128 << 16) | (19 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_UART2  = (1 << 30) | (2 << 26) | (192 << 16) | (18 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_UART1  = (1 << 30) | (3 << 26) | (256 << 16) | (17 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_UART0  = (1 << 30) | (4 << 26) | (320 << 16) | (16 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_SSP1   = (1 << 30) | (5 << 26) | (384 << 16) | (15 << 11) | (1 << 10),
    CLKPWR_BASECLOCK_SSP0   = (1 << 30) | (6 << 26) | (448 << 16) | (14 << 11) | (1 << 10),
} CLKPWR_BaseClock;


typedef enum CLKPWR_Clockswitch {
    /* 30-11:  Base clock info
     * 10-0:   Register offset in CCUx
     */
    CLKPWR_CLOCKSWITCH_I2C1         =  66 | CLKPWR_BASECLOCK_APB3,
    CLKPWR_CLOCKSWITCH_DAC          =  68 | CLKPWR_BASECLOCK_APB3,
    CLKPWR_CLOCKSWITCH_ADC0         =  70 | CLKPWR_BASECLOCK_APB3,
    CLKPWR_CLOCKSWITCH_ADC1         =  72 | CLKPWR_BASECLOCK_APB3,
    CLKPWR_CLOCKSWITCH_CAN0         =  74 | CLKPWR_BASECLOCK_APB3,

    CLKPWR_CLOCKSWITCH_MOTOCON_PWM  = 130 | CLKPWR_BASECLOCK_APB1,
    CLKPWR_CLOCKSWITCH_I2C0         = 132 | CLKPWR_BASECLOCK_APB1,
    CLKPWR_CLOCKSWITCH_I2S          = 134 | CLKPWR_BASECLOCK_APB1,
    CLKPWR_CLOCKSWITCH_CAN1         = 136 | CLKPWR_BASECLOCK_APB1,

    CLKPWR_CLOCKSWITCH_SPIFI_CLKI   = 192 | CLKPWR_BASECLOCK_SPIFI,

    CLKPWR_CLOCKSWITCH_SPIFI        = 258 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_GPIO         = 260 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_LCD          = 262 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_ETHERNET     = 264 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_USB0         = 266 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_EMC          = 268 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_SDIO         = 270 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_DMA          = 272 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_CPU          = 274 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_AES          = 280 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_SCT          = 282 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_USB1         = 284 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_EMCDIV       = 286 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_WWDT         = 320 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_USART0       = 322 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_UART1        = 324 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_SSP0         = 326 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_TIMER0       = 328 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_TIMER1       = 330 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_SCU          = 332 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_CREG         = 334 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_RITIMER      = 384 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_USART2       = 386 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_USART3       = 388 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_TIMER2       = 390 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_TIMER3       = 392 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_SSP1         = 394 | CLKPWR_BASECLOCK_M3,
    CLKPWR_CLOCKSWITCH_QEI          = 396 | CLKPWR_BASECLOCK_M3,
} CLKPWR_Clockswitch;


typedef enum CLKPWR_Clock {
    CLKPWR_CLOCK_I2C1               = CLKPWR_CLOCKSWITCH_I2C1,
    CLKPWR_CLOCK_DAC                = CLKPWR_CLOCKSWITCH_DAC,
    CLKPWR_CLOCK_ADC0               = CLKPWR_CLOCKSWITCH_ADC0,
    CLKPWR_CLOCK_ADC1               = CLKPWR_CLOCKSWITCH_ADC1,
    CLKPWR_CLOCK_CAN0               = CLKPWR_CLOCKSWITCH_CAN0,

    CLKPWR_CLOCK_MOTOCON_PWM        = CLKPWR_CLOCKSWITCH_MOTOCON_PWM,
    CLKPWR_CLOCK_I2C0               = CLKPWR_CLOCKSWITCH_I2C0,
    CLKPWR_CLOCK_I2S                = CLKPWR_CLOCKSWITCH_I2S,
    CLKPWR_CLOCK_CAN1               = CLKPWR_CLOCKSWITCH_CAN1,

    CLKPWR_CLOCK_SPIFI_CLKI         = CLKPWR_CLOCKSWITCH_SPIFI_CLKI,
                
    CLKPWR_CLOCK_SPIFI              = CLKPWR_CLOCKSWITCH_SPIFI,
    CLKPWR_CLOCK_GPIO               = CLKPWR_CLOCKSWITCH_GPIO,
    CLKPWR_CLOCK_LCD                = CLKPWR_CLOCKSWITCH_LCD,
    CLKPWR_CLOCK_ETHERNET           = CLKPWR_CLOCKSWITCH_ETHERNET,
    CLKPWR_CLOCK_USB0               = CLKPWR_CLOCKSWITCH_USB0,
    CLKPWR_CLOCK_EMC                = CLKPWR_CLOCKSWITCH_EMC,
    CLKPWR_CLOCK_SDIO               = CLKPWR_CLOCKSWITCH_SDIO,
    CLKPWR_CLOCK_DMA                = CLKPWR_CLOCKSWITCH_DMA,
    CLKPWR_CLOCK_CPU                = CLKPWR_CLOCKSWITCH_CPU,
    CLKPWR_CLOCK_AES                = CLKPWR_CLOCKSWITCH_AES,
    CLKPWR_CLOCK_SCT                = CLKPWR_CLOCKSWITCH_SCT,
    CLKPWR_CLOCK_USB1               = CLKPWR_CLOCKSWITCH_USB1,
    CLKPWR_CLOCK_EMCDIV             = CLKPWR_CLOCKSWITCH_EMCDIV,
    CLKPWR_CLOCK_WWDT               = CLKPWR_CLOCKSWITCH_WWDT,
    CLKPWR_CLOCK_USART0             = CLKPWR_CLOCKSWITCH_USART0,
    CLKPWR_CLOCK_UART1              = CLKPWR_CLOCKSWITCH_UART1,
    CLKPWR_CLOCK_SSP0               = CLKPWR_CLOCKSWITCH_SSP0,
    CLKPWR_CLOCK_TIMER0             = CLKPWR_CLOCKSWITCH_TIMER0,
    CLKPWR_CLOCK_TIMER1             = CLKPWR_CLOCKSWITCH_TIMER1,
    CLKPWR_CLOCK_SCU                = CLKPWR_CLOCKSWITCH_SCU,
    CLKPWR_CLOCK_CREG               = CLKPWR_CLOCKSWITCH_CREG,
    CLKPWR_CLOCK_RITIMER            = CLKPWR_CLOCKSWITCH_RITIMER,
    CLKPWR_CLOCK_USART2             = CLKPWR_CLOCKSWITCH_USART2,
    CLKPWR_CLOCK_USART3             = CLKPWR_CLOCKSWITCH_USART3,
    CLKPWR_CLOCK_TIMER2             = CLKPWR_CLOCKSWITCH_TIMER2,
    CLKPWR_CLOCK_TIMER3             = CLKPWR_CLOCKSWITCH_TIMER3,
    CLKPWR_CLOCK_SSP1               = CLKPWR_CLOCKSWITCH_SSP1,
    CLKPWR_CLOCK_QEI                = CLKPWR_CLOCKSWITCH_QEI,
} CLKPWR_Clock;

#if LPCLIB_FAMILY == LPCLIB_FAMILY_LPC178X
typedef enum CLKPWR_Reset {
    CLKPWR_RESET_LCD = 0,
    CLKPWR_RESET_TIM0 = 1,
    CLKPWR_RESET_TIM1 = 2,
    CLKPWR_RESET_UART0 = 3,
    CLKPWR_RESET_UART1 = 4,
    CLKPWR_RESET_PWM0 = 5,
    CLKPWR_RESET_PWM1 = 6,
    CLKPWR_RESET_I2C0 = 7,
    CLKPWR_RESET_UART4 = 8,
    CLKPWR_RESET_RTC = 9,
    CLKPWR_RESET_SSP1 = 10,
    CLKPWR_RESET_EMC = 11,
    CLKPWR_RESET_ADC = 12,
    CLKPWR_RESET_CAN1 = 13,
    CLKPWR_RESET_CAN2 = 14,
    CLKPWR_RESET_GPIO = 15,
    CLKPWR_RESET_MCPWM = 17,
    CLKPWR_RESET_QEI = 18,
    CLKPWR_RESET_I2C1 = 19,
    CLKPWR_RESET_SSP2 = 20,
    CLKPWR_RESET_SSP0 = 21,
    CLKPWR_RESET_TIM2 = 22,
    CLKPWR_RESET_TIM3 = 23,
    CLKPWR_RESET_UART2 = 24,
    CLKPWR_RESET_UART3 = 25,
    CLKPWR_RESET_I2C2 = 26,
    CLKPWR_RESET_I2S = 27,
    CLKPWR_RESET_SDC = 28,
    CLKPWR_RESET_GPDMA = 29,
    CLKPWR_RESET_ENET = 30,
    CLKPWR_RESET_USB = 31,
    CLKPWR_RESET_IOCON = 32 + 0,
    CLKPWR_RESET_DAC = 32 + 1,
    CLKPWR_RESET_CANACC = 32 + 2,
} CLKPWR_Reset;
#endif

typedef enum CLKPWR_UnitPower {
    CLKPWR_UNITPOWER___DUMMY__,
} CLKPWR_UnitPower;



/** @} CLKPWR Types, enums, macros */


/** \defgroup CLKPWR_Public_Functions CLKPWR API Functions
 *  @{
 */


/** Return the clock frequency of an internal bus.
 *
 *  \param[in] bus The bus to be queried
 *  \return Bus clock in Hz
 */
uint32_t CLKPWR_getBusClock (CLKPWR_Clock clock);


/** Enable a clock signal.
 *
 *  \param[in] clock Clock switch selector
 */
static void CLKPWR_enableClock (CLKPWR_Clockswitch clock);


/** Disable a clock signal.
 *
 *  \param[in] clock Clock switch selector
 */
static void CLKPWR_disableClock (CLKPWR_Clockswitch clock);


/** Power up a functional unit.
 *
 *  \param[in] unit Unit selector
 */
static void CLKPWR_powerUpUnit (CLKPWR_UnitPower unit);


/** Power down a functional unit.
 *
 *  \param[in] unit Unit selector
 */
static void CLKPWR_powerDownUnit (CLKPWR_UnitPower unit);


#if LPCLIB_FAMILY == LPCLIB_FAMILY_LPC178X
/** Assert the reset signal of the selected peripheral block.
 *
 *  \param[in] peripheral Peripheral block selector
 */
static void CLKPWR_assertPeripheralReset (CLKPWR_Reset peripheral);
#endif


#if LPCLIB_FAMILY == LPCLIB_FAMILY_LPC178X
/** Deassert the reset signal of the selected peripheral block.
 *
 *  \param[in] peripheral Peripheral block selector
 */
static void CLKPWR_deassertPeripheralReset (CLKPWR_Reset peripheral);
#endif


/** Attempts to set the CPU clock to the desired frequency.
 *
 *  Sets the CPU clock to the frequency which is closest to the requested
 *  value. Assumes that the PLL input clock has been selected before.
 *
 *  \param[in] targetCpuFrequency Desired CPU frequency (Hz)
 *  \return ...
 */
LPCLIB_Result CLKPWR_setCpuClock (uint32_t targetCpuFrequency);



/** Sets up the USB clock using the USB PLL.
 *
 *  Only works if the crystal frequency is an integer fraction of 48 MHz.
 *
 *  \param[in] targetCpuFrequency Desired CPU frequency (kHz)
 *  \param[out] newCpuFrequency CPU frequency (kHz) after the call
 *  \return ...
 */
LPCLIB_Result CLKPWR_setUsbClock (void);


/** Select an input for a base clock.
 *
 *  Connect a base clock to the specified input clock.
 *
 *  \param[in] baseClock Base clock
 *  \param[in] sourceClock Input clock
 *  \retval LPCLIB_SUCCESS Ok
 */
LPCLIB_Result CLKPWR_connectBaseClock (CLKPWR_BaseClock baseClock, CLKPWR_Source sourceClock);


/** Enable the FPU.
 *
 *  Enable the FPU (on devices that implement it).
 */
void CLKPWR_enableFPU (void);



__FORCEINLINE(void CLKPWR_powerUpUnit (CLKPWR_UnitPower unit))
{
    (void) unit;
}

__FORCEINLINE(void CLKPWR_powerDownUnit (CLKPWR_UnitPower unit))
{
    (void) unit;
}

#if LPCLIB_FAMILY == LPCLIB_FAMILY_LPC178X
__FORCEINLINE(void CLKPWR_assertPeripheralReset (CLKPWR_Reset peripheral))
{
    /* Note: The parameter "peripheral" can have values from 0...63.
     *       Yet the bitband access works since the registers RSTCON0 and RSTCON1
     *       are on consecutive addresses.
     */
}
#endif

#if LPCLIB_FAMILY == LPCLIB_FAMILY_LPC178X
__FORCEINLINE(void CLKPWR_deassertPeripheralReset (const CLKPWR_Reset peripheral))
{
    /* Note: The parameter "peripheral" can have values from 0...63.
     *       Yet the bitband access works since the registers RSTCON0 and RSTCON1
     *       are on consecutive addresses.
     */
}
#endif


__FORCEINLINE(void CLKPWR_enableClock (CLKPWR_Clockswitch clock))
{
    (void)clock;
}



__FORCEINLINE(void CLKPWR_disableClock (CLKPWR_Clockswitch clock))
{
    (void)clock;
}





/** @} CLKPWR API Functions */

/** @} CLKPWR */

#endif /* #ifndef __LPC43XX_CLKPWR_H__ */

