/* $Id::                                                               $
 *
 * Copyright (c) 2011, NXP Semiconductors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file
 *  \brief IOCON driver interface.
 *
 *  This file contains the API for the IOCON block.
 *
 *  \author NXP Semiconductors
 */


#ifndef __LPC43XX_IOCON_H__
#define __LPC43XX_IOCON_H__

#ifdef __cplusplus
extern "C"
{
#endif

/** \defgroup IOCON
 *  \ingroup API
 *  @{
 */

#include "lpc43xx_libconfig.h"
#include "lpclib_types.h"
#include "lpc43xx_clkpwr.h"


/** \defgroup IOCON_Public_Types IOCON Types, enums, macros
 *  @{
 */


/** Pin classification based on their options and electrical characteristics. */
enum IOCON_PinType {
    IOCON_PINTYPE_D  = (0u << 28),                      /**< Normal drive digital pin */
    IOCON_PINTYPE_H  = (1u << 28),                      /**< High drive digital pin */
    IOCON_PINTYPE_A0 = IOCON_PINTYPE_D | (2u << 24),    /**< ADC0 pin */
    IOCON_PINTYPE_A1 = IOCON_PINTYPE_D | (3u << 24),    /**< ADC1 pin */
    IOCON_PINTYPE_A  = IOCON_PINTYPE_D | (4u << 24),    /**< Analog pin */
    IOCON_PINTYPE_U  = (2u << 28),                      /**< USB1 pins */
    IOCON_PINTYPE_I  = (3u << 28),                      /**< I2C pins */
    IOCON_PINTYPE_C  = (4u << 28),                      /**< Clock pins */
};


typedef enum IOCON_PinName {
    PIN_P0_0  =   0 | IOCON_PINTYPE_D,
    PIN_P0_1  =   1 | IOCON_PINTYPE_D,
    PIN_P1_0  =  32 | IOCON_PINTYPE_D,
    PIN_P1_1  =  33 | IOCON_PINTYPE_D,
    PIN_P1_2  =  34 | IOCON_PINTYPE_D,
    PIN_P1_3  =  35 | IOCON_PINTYPE_D,
    PIN_P1_4  =  36 | IOCON_PINTYPE_D,
    PIN_P1_5  =  37 | IOCON_PINTYPE_D,
    PIN_P1_6  =  38 | IOCON_PINTYPE_D,
    PIN_P1_7  =  39 | IOCON_PINTYPE_D,
    PIN_P1_8  =  40 | IOCON_PINTYPE_D,
    PIN_P1_9  =  41 | IOCON_PINTYPE_D,
    PIN_P1_10 =  42 | IOCON_PINTYPE_D,
    PIN_P1_11 =  43 | IOCON_PINTYPE_D,
    PIN_P1_12 =  44 | IOCON_PINTYPE_D,
    PIN_P1_13 =  45 | IOCON_PINTYPE_D,
    PIN_P1_14 =  46 | IOCON_PINTYPE_D,
    PIN_P1_15 =  47 | IOCON_PINTYPE_D,
    PIN_P1_16 =  48 | IOCON_PINTYPE_D,
    PIN_P1_17 =  49 | IOCON_PINTYPE_H,
    PIN_P1_18 =  50 | IOCON_PINTYPE_D,
    PIN_P1_19 =  51 | IOCON_PINTYPE_D,
    PIN_P1_20 =  52 | IOCON_PINTYPE_D,
    PIN_P2_0  =  64 | IOCON_PINTYPE_D,
    PIN_P2_1  =  65 | IOCON_PINTYPE_D,
    PIN_P2_2  =  66 | IOCON_PINTYPE_D,
    PIN_P2_3  =  67 | IOCON_PINTYPE_H,
    PIN_P2_4  =  68 | IOCON_PINTYPE_H,
    PIN_P2_5  =  69 | IOCON_PINTYPE_H,
    PIN_P2_6  =  70 | IOCON_PINTYPE_D,
    PIN_P2_7  =  71 | IOCON_PINTYPE_D,
    PIN_P2_8  =  72 | IOCON_PINTYPE_D,
    PIN_P2_9  =  73 | IOCON_PINTYPE_D,
    PIN_P2_10 =  74 | IOCON_PINTYPE_D,
    PIN_P2_11 =  75 | IOCON_PINTYPE_D,
    PIN_P2_12 =  76 | IOCON_PINTYPE_D,
    PIN_P2_13 =  77 | IOCON_PINTYPE_D,
    PIN_P3_0  =  96 | IOCON_PINTYPE_D,
    PIN_P3_1  =  97 | IOCON_PINTYPE_D,
    PIN_P3_2  =  98 | IOCON_PINTYPE_D,
    PIN_P3_3  =  99 | IOCON_PINTYPE_D,
    PIN_P3_4  = 100 | IOCON_PINTYPE_D,
    PIN_P3_5  = 101 | IOCON_PINTYPE_D,
    PIN_P3_6  = 102 | IOCON_PINTYPE_D,
    PIN_P3_7  = 103 | IOCON_PINTYPE_D,
    PIN_P3_8  = 104 | IOCON_PINTYPE_D,
    PIN_P4_0  = 128 | IOCON_PINTYPE_D,
    PIN_P4_1  = 129 | IOCON_PINTYPE_A0 | (1u << 16),
    PIN_P4_2  = 130 | IOCON_PINTYPE_D,
    PIN_P4_3  = 131 | IOCON_PINTYPE_A0 | (0u << 16),
    PIN_P4_4  = 132 | IOCON_PINTYPE_A,
    PIN_P4_5  = 133 | IOCON_PINTYPE_D,
    PIN_P4_6  = 134 | IOCON_PINTYPE_D,
    PIN_P4_7  = 135 | IOCON_PINTYPE_D,
    PIN_P4_8  = 136 | IOCON_PINTYPE_D,
    PIN_P4_9  = 137 | IOCON_PINTYPE_D,
    PIN_P4_10 = 138 | IOCON_PINTYPE_D,
    PIN_P5_0  = 160 | IOCON_PINTYPE_D,
    PIN_P5_1  = 161 | IOCON_PINTYPE_D,
    PIN_P5_2  = 162 | IOCON_PINTYPE_D,
    PIN_P5_3  = 163 | IOCON_PINTYPE_D,
    PIN_P5_4  = 164 | IOCON_PINTYPE_D,
    PIN_P5_5  = 165 | IOCON_PINTYPE_D,
    PIN_P5_6  = 166 | IOCON_PINTYPE_D,
    PIN_P5_7  = 167 | IOCON_PINTYPE_D,
    PIN_P6_0  = 192 | IOCON_PINTYPE_D,
    PIN_P6_1  = 193 | IOCON_PINTYPE_D,
    PIN_P6_2  = 194 | IOCON_PINTYPE_D,
    PIN_P6_3  = 195 | IOCON_PINTYPE_D,
    PIN_P6_4  = 196 | IOCON_PINTYPE_D,
    PIN_P6_5  = 197 | IOCON_PINTYPE_D,
    PIN_P6_6  = 198 | IOCON_PINTYPE_D,
    PIN_P6_7  = 199 | IOCON_PINTYPE_D,
    PIN_P6_8  = 200 | IOCON_PINTYPE_D,
    PIN_P6_9  = 201 | IOCON_PINTYPE_D,
    PIN_P6_10 = 202 | IOCON_PINTYPE_D,
    PIN_P6_11 = 203 | IOCON_PINTYPE_D,
    PIN_P6_12 = 204 | IOCON_PINTYPE_D,
    PIN_P7_0  = 224 | IOCON_PINTYPE_D,
    PIN_P7_1  = 225 | IOCON_PINTYPE_D,
    PIN_P7_2  = 226 | IOCON_PINTYPE_D,
    PIN_P7_3  = 227 | IOCON_PINTYPE_D,
    PIN_P7_4  = 228 | IOCON_PINTYPE_A0 | (4u << 16),
    PIN_P7_5  = 229 | IOCON_PINTYPE_A0 | (3u << 16),
    PIN_P7_6  = 230 | IOCON_PINTYPE_D,
    PIN_P7_7  = 231 | IOCON_PINTYPE_A1 | (6u << 16),
    PIN_P8_0  = 256 | IOCON_PINTYPE_H,
    PIN_P8_1  = 257 | IOCON_PINTYPE_H,
    PIN_P8_2  = 258 | IOCON_PINTYPE_H,
    PIN_P8_3  = 259 | IOCON_PINTYPE_D,
    PIN_P8_4  = 260 | IOCON_PINTYPE_D,
    PIN_P8_5  = 261 | IOCON_PINTYPE_D,
    PIN_P8_6  = 262 | IOCON_PINTYPE_D,
    PIN_P8_7  = 263 | IOCON_PINTYPE_D,
    PIN_P8_8  = 264 | IOCON_PINTYPE_D,
    PIN_P9_0  = 288 | IOCON_PINTYPE_D,
    PIN_P9_1  = 289 | IOCON_PINTYPE_D,
    PIN_P9_2  = 290 | IOCON_PINTYPE_D,
    PIN_P9_3  = 291 | IOCON_PINTYPE_D,
    PIN_P9_4  = 292 | IOCON_PINTYPE_D,
    PIN_P9_5  = 293 | IOCON_PINTYPE_D,
    PIN_P9_6  = 294 | IOCON_PINTYPE_D,
    PIN_PA_0  = 320 | IOCON_PINTYPE_D,
    PIN_PA_1  = 321 | IOCON_PINTYPE_H,
    PIN_PA_2  = 322 | IOCON_PINTYPE_H,
    PIN_PA_3  = 323 | IOCON_PINTYPE_H,
    PIN_PA_4  = 324 | IOCON_PINTYPE_D,
    PIN_PB_0  = 352 | IOCON_PINTYPE_D,
    PIN_PB_1  = 353 | IOCON_PINTYPE_D,
    PIN_PB_2  = 354 | IOCON_PINTYPE_D,
    PIN_PB_3  = 355 | IOCON_PINTYPE_D,
    PIN_PB_4  = 356 | IOCON_PINTYPE_D,
    PIN_PB_5  = 357 | IOCON_PINTYPE_D,
    PIN_PB_6  = 358 | IOCON_PINTYPE_A0 | (6u << 16),
    PIN_PC_0  = 384 | IOCON_PINTYPE_A1 | (1u << 16),
    PIN_PC_1  = 385 | IOCON_PINTYPE_D,
    PIN_PC_2  = 386 | IOCON_PINTYPE_D,
    PIN_PC_3  = 387 | IOCON_PINTYPE_A1 | (0u << 16),
    PIN_PC_4  = 388 | IOCON_PINTYPE_D,
    PIN_PC_5  = 389 | IOCON_PINTYPE_D,
    PIN_PC_6  = 390 | IOCON_PINTYPE_D,
    PIN_PC_7  = 391 | IOCON_PINTYPE_D,
    PIN_PC_8  = 392 | IOCON_PINTYPE_D,
    PIN_PC_9  = 393 | IOCON_PINTYPE_D,
    PIN_PC_10 = 394 | IOCON_PINTYPE_D,
    PIN_PC_11 = 395 | IOCON_PINTYPE_D,
    PIN_PC_12 = 396 | IOCON_PINTYPE_D,
    PIN_PC_13 = 397 | IOCON_PINTYPE_D,
    PIN_PC_14 = 398 | IOCON_PINTYPE_D,
    PIN_PD_0  = 416 | IOCON_PINTYPE_D,
    PIN_PD_1  = 417 | IOCON_PINTYPE_D,
    PIN_PD_2  = 418 | IOCON_PINTYPE_D,
    PIN_PD_3  = 419 | IOCON_PINTYPE_D,
    PIN_PD_4  = 420 | IOCON_PINTYPE_D,
    PIN_PD_5  = 421 | IOCON_PINTYPE_D,
    PIN_PD_6  = 422 | IOCON_PINTYPE_D,
    PIN_PD_7  = 423 | IOCON_PINTYPE_D,
    PIN_PD_8  = 424 | IOCON_PINTYPE_D,
    PIN_PD_9  = 425 | IOCON_PINTYPE_D,
    PIN_PD_10 = 426 | IOCON_PINTYPE_D,
    PIN_PD_11 = 427 | IOCON_PINTYPE_D,
    PIN_PD_12 = 428 | IOCON_PINTYPE_D,
    PIN_PD_13 = 429 | IOCON_PINTYPE_D,
    PIN_PD_14 = 430 | IOCON_PINTYPE_D,
    PIN_PD_15 = 431 | IOCON_PINTYPE_D,
    PIN_PD_16 = 432 | IOCON_PINTYPE_D,
    PIN_PE_0  = 448 | IOCON_PINTYPE_D,
    PIN_PE_1  = 449 | IOCON_PINTYPE_D,
    PIN_PE_2  = 450 | IOCON_PINTYPE_D,
    PIN_PE_3  = 451 | IOCON_PINTYPE_D,
    PIN_PE_4  = 452 | IOCON_PINTYPE_D,
    PIN_PE_5  = 453 | IOCON_PINTYPE_D,
    PIN_PE_6  = 454 | IOCON_PINTYPE_D,
    PIN_PE_7  = 455 | IOCON_PINTYPE_D,
    PIN_PE_8  = 456 | IOCON_PINTYPE_D,
    PIN_PE_9  = 457 | IOCON_PINTYPE_D,
    PIN_PE_10 = 458 | IOCON_PINTYPE_D,
    PIN_PE_11 = 459 | IOCON_PINTYPE_D,
    PIN_PE_12 = 460 | IOCON_PINTYPE_D,
    PIN_PE_13 = 461 | IOCON_PINTYPE_D,
    PIN_PE_14 = 462 | IOCON_PINTYPE_D,
    PIN_PE_15 = 463 | IOCON_PINTYPE_D,
    PIN_PF_0  = 480 | IOCON_PINTYPE_D,
    PIN_PF_1  = 481 | IOCON_PINTYPE_D,
    PIN_PF_2  = 482 | IOCON_PINTYPE_D,
    PIN_PF_3  = 483 | IOCON_PINTYPE_D,
    PIN_PF_4  = 484 | IOCON_PINTYPE_D,
    PIN_PF_5  = 485 | IOCON_PINTYPE_A1 | (4u << 16),
    PIN_PF_6  = 486 | IOCON_PINTYPE_A1 | (3u << 16),
    PIN_PF_7  = 487 | IOCON_PINTYPE_A1 | (7u << 16),
    PIN_PF_8  = 488 | IOCON_PINTYPE_A0 | (2u << 16),
    PIN_PF_9  = 489 | IOCON_PINTYPE_A1 | (2u << 16),
    PIN_PF_10 = 490 | IOCON_PINTYPE_A0 | (5u << 16),
    PIN_PF_11 = 491 | IOCON_PINTYPE_A1 | (5u << 16),

    PIN_CLK0  = 768 | IOCON_PINTYPE_C,
    PIN_CLK1  = 769 | IOCON_PINTYPE_C,
    PIN_CLK2  = 770 | IOCON_PINTYPE_C,
    PIN_CLK3  = 771 | IOCON_PINTYPE_C,
    PIN_I2C0  = 801 | IOCON_PINTYPE_I,
} IOCON_PinName;


typedef enum IOCON_Function {
    PIN_FUNCTION_0 = 0,         /**< Function selector 0 */
    PIN_FUNCTION_1 = 1,         /**< Function selector 1 */
    PIN_FUNCTION_2 = 2,         /**< Function selector 2 */
    PIN_FUNCTION_3 = 3,         /**< Function selector 3 */
    PIN_FUNCTION_4 = 4,         /**< Function selector 4 */
    PIN_FUNCTION_5 = 5,         /**< Function selector 5 */
    PIN_FUNCTION_6 = 6,         /**< Function selector 6 */
    PIN_FUNCTION_7 = 7,         /**< Function selector 7 */
} IOCON_Function;


typedef enum IOCON_Pull {
    PIN_PULL_NONE = 2,          /**< No pull-up or pull-down resistor */
    PIN_PULL_DOWN = 3,          /**< Internal pull-down resistor */
    PIN_PULL_UP   = 0,          /**< Internal pull-up resistor */
    PIN_PULL_REPEATER = 1,      /**< Keep last actively driven input level */
} IOCON_Pull;

typedef enum IOCON_SlewRate {
    PIN_SLEWRATE_SLOW = 0,      /**< Slow slew rate */
    PIN_SLEWRATE_FAST  = 1,     /**< Fast slew rate */
} IOCON_SlewRate;

typedef enum IOCON_InputBuffer {
    PIN_INPUTBUFFER_OFF = 0,    /**< Input buffer disabled */
    PIN_INPUTBUFFER_ON  = 1,    /**< Input buffer enabled */
} IOCON_InputBuffer;


typedef enum IOCON_DriveStrength {
    PIN_DRIVESTRENGTH_4MA = 0,  /**< 4 mA driver */
    PIN_DRIVESTRENGTH_8MA = 1,  /**< 8 mA driver */
    PIN_DRIVESTRENGTH_14MA = 2, /**< 14 mA driver */
    PIN_DRIVESTRENGTH_20MA = 3, /**< 20 mA driver */
} IOCON_DriveStrength;


typedef enum IOCON_I2c0GlitchFilter {
    PIN_I2C0_GLITCHFILTER_NONE = 0x8080,    /**< No glitch filter */
    PIN_I2C0_GLITCHFILTER_3NS = 0x0101,     /**< Glitch filter enabled, 3 ns */
    PIN_I2C0_GLITCHFILTER_50NS = 0x0000,    /**< Glitch filter enabled, 50 ns */
} IOCON_I2c0GlitchFilter;

typedef enum IOCON_I2c0Speed {
    PIN_I2C0_SPEED_STANDARD = 0x0000,       /**< No glitch filter */
    PIN_I2C0_SPEED_FASTMODE = 0x0000,       /**< No glitch filter */
    PIN_I2C0_SPEED_FASTMODEPLUS = 0x0404,   /**< No glitch filter */
} IOCON_I2c0Speed;


typedef enum IOCON_InputGlitchFilter {
    PIN_INPUT_GLITCHFILTER_OFF = 1,         /**< No glitch filter (> 30 MHz) */
    PIN_INPUT_GLITCHFILTER_ON = 0,          /**< With glitch filter (<= 30 MHz) */
} IOCON_InputGlitchFilter;


typedef uint32_t IOCON_PinConfig;



/** @} IOCON Types enums, macros */


/** \defgroup IOCON_Public_Functions IOCON API Functions
 *  @{
 */


/** Initialize the IOCON module.
 */
void IOCON_open (void);


/** Configure a pin.
 *
 *  \param[in] pin Selects the pin to be configured
 *  \param[in] config Provides the configuration (as obtained by IOCON_makeConfigX())
 */
static void IOCON_configurePin (const IOCON_PinName pin, const IOCON_PinConfig config);


/** Helper function to create a pin configuration from the settings.
 *
 *  \param[in] function Selects pin function (GPIO, etc.)
 *  \param[in] pull Pull-up/pull-down selection
 *  \param[in] slewRate Slow/fast slew rate
 *  \param[in] inputBuffer Enable/disable input buffer
 *  \param[in] glitchFilter 30 MHz glitch filter
 *  \return All settings combined in one word
 */
static IOCON_PinConfig IOCON_makeConfigD (
        IOCON_Function function,
        IOCON_Pull pull,
        IOCON_SlewRate slewRate,
        IOCON_InputBuffer inputBuffer,
        IOCON_InputGlitchFilter glitchFilter);


/** Helper function to create a pin configuration from the settings.
 *
 *  \param[in] function Selects pin function (GPIO, etc.)
 *  \param[in] pull Pull-up/pull-down selection
 *  \param[in] slewRate Slow/fast slew rate
 *  \param[in] inputBuffer Enable/disable input buffer
 *  \param[in] glitchFilter 30 MHz glitch filter
 *  \param[in] driveStrength Current capability of pin
 *  \return All settings combined in one word
 */
static IOCON_PinConfig IOCON_makeConfigH (
        IOCON_Function function,
        IOCON_Pull pull,
        IOCON_SlewRate slewRate,
        IOCON_InputBuffer inputBuffer,
        IOCON_InputGlitchFilter glitchFilter,
        IOCON_DriveStrength driveStrength);


/** Helper function to create a pin configuration from the settings.
 *
 *  \param[in] glitchFilter Selects glitch filter time constant
 *  \param[in] speed Selects Standard/FastMode/FastMode+
 *  \return All settings combined in one word
 */
static IOCON_PinConfig IOCON_makeConfigI (
        IOCON_I2c0GlitchFilter glitchFilter,
        IOCON_I2c0Speed speed);


/** Helper function to create a pin configuration from the settings.
 *
 *  \param[in] function Selects pin function (GPIO, etc.)
 *  \param[in] pull Pull-up/pull-down selection
 *  \param[in] slewRate Slow/fast slew rate
 *  \param[in] inputBuffer Enable/disable input buffer
 *  \param[in] glitchFilter 30 MHz glitch filter
 *  \return All settings combined in one word
 */
static IOCON_PinConfig IOCON_makeConfigC (
        IOCON_Function function,
        IOCON_Pull pull,
        IOCON_SlewRate slewRate,
        IOCON_InputBuffer inputBuffer,
        IOCON_InputGlitchFilter glitchFilter);


/** Check if an error occured during any of the pin configurations done so far.
 *
 *  \retval LPCLIB_SUCCESS No problems occurred.
 *  \retval LPCLIB_ILLEGAL_PARAMETER At least one mismatch between pin and
 *          makeConfigX function found
 */
static LPCLIB_Result IOCON_checkErrors (void);

/** @} IOCON API Functions */



__FORCEINLINE(IOCON_PinConfig IOCON_makeConfigD (
        IOCON_Function function,
        IOCON_Pull pull,
        IOCON_SlewRate slewRate,
        IOCON_InputBuffer inputBuffer,
        IOCON_InputGlitchFilter glitchFilter))
{
    return (IOCON_PinConfig)(
        (function << 0)     |
        (pull << 3)         |
        (slewRate << 5)     |
        (inputBuffer << 6)  |
        (glitchFilter << 7) |
        IOCON_PINTYPE_D
        );
}


__FORCEINLINE(IOCON_PinConfig IOCON_makeConfigH (
        IOCON_Function function,
        IOCON_Pull pull,
        IOCON_SlewRate slewRate,
        IOCON_InputBuffer inputBuffer,
        IOCON_InputGlitchFilter glitchFilter,
        IOCON_DriveStrength driveStrength))
{
    return (IOCON_PinConfig)(
        (function << 0)     |
        (pull << 3)         |
        (slewRate << 5)     |
        (inputBuffer << 6)  |
        (glitchFilter << 7) |
        (driveStrength << 8)|
        IOCON_PINTYPE_H
        );
}


__FORCEINLINE(IOCON_PinConfig IOCON_makeConfigI (
        IOCON_I2c0GlitchFilter glitchFilter,
        IOCON_I2c0Speed speed))
{
    return (IOCON_PinConfig)(
        (glitchFilter)      |
        (speed)             |
        0x00000808          |
        IOCON_PINTYPE_I
        );
}


__FORCEINLINE(IOCON_PinConfig IOCON_makeConfigC (
        IOCON_Function function,
        IOCON_Pull pull,
        IOCON_SlewRate slewRate,
        IOCON_InputBuffer inputBuffer,
        IOCON_InputGlitchFilter glitchFilter))
{
    return (IOCON_PinConfig)(
        (function << 0)     |
        (pull << 3)         |
        (slewRate << 5)     |
        (inputBuffer << 6)  |
        (glitchFilter << 7) |
        IOCON_PINTYPE_C
        );
}


extern LPCLIB_Result __IOCON_configError;

#define __ENAI0_BITBAND     ((volatile uint32_t *)0x430D9100)
#define __ENAI1_BITBAND     ((volatile uint32_t *)0x430D9180)

__FORCEINLINE(void IOCON_configurePin (const IOCON_PinName pin, const IOCON_PinConfig config))
{
    if ((pin >> 28) != (config >> 28)) {                /* Type encoded into the upper 4 bits */
        __IOCON_configError = LPCLIB_ILLEGAL_PARAMETER;
        return;
    }

    if ((pin & 0xFF000000) == IOCON_PINTYPE_A0) {
        if ((pin & 0xFFFF) == 8) {
            __ENAI0_BITBAND[(pin >> 16) & 0x0F] = 1;
        }
        else {
            __ENAI0_BITBAND[(pin >> 16) & 0x0F] = 0;
        }
    }

    if ((pin & 0xFF000000) == IOCON_PINTYPE_A1) {
        if ((pin & 0xFFFF) == 8) {
            __ENAI1_BITBAND[(pin >> 16) & 0x0F] = 1;
        }
        else {
            __ENAI1_BITBAND[(pin >> 16) & 0x0F] = 0;
        }
    }

    ((volatile uint32_t *)LPC_SCU)[pin & 0xFFFF] = (uint32_t)config & 0x0FFFFFFF;
}


__FORCEINLINE(LPCLIB_Result IOCON_checkErrors (void))
{
    return __IOCON_configError;
}


/** @} IOCON API Functions */

/** @} IOCON */

#ifdef __cplusplus
}
#endif

#endif /* #ifndef __LPC43xx_IOCON_H__ */

