/* $Id::                                                               $
 *
 * Copyright (c) 2011, NXP Semiconductors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file
 *  \brief CLKPWR driver implementation.
 *
 *  This file contains the implementation of the CLKPWR block driver.
 *
 *  \author NXP Semiconductors
 */

/** \addtogroup CLKPWR
 *  @{
 */

#include "lpc43xx_libconfig.h"


#include "system_LPC43xx.h"
#include "lpc43xx_clkpwr.h"


LPCLIB_DefineRegBit(CGU_XTAL_OSC_CTRL_ENABLE,   0,  1);
LPCLIB_DefineRegBit(CGU_XTAL_OSC_CTRL_HF,       2,  1);

LPCLIB_DefineRegBit(CGU_OUTCLK_PD,              0,  1);
LPCLIB_DefineRegBit(CGU_OUTCLK_AUTOBLOCK,       11, 1);
LPCLIB_DefineRegBit(CGU_OUTCLK_CLK_SEL,         24, 5);

LPCLIB_DefineRegBit(CGU_PLL1_STAT_LOCK,         0,  1);

LPCLIB_DefineRegBit(CGU_PLL1_CTRL_PD,           0,  1);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_BYPASS,       1,  1);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_FBSEL,        6,  1);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_DIRECT,       7,  1);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_PSEL,         8,  2);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_AUTOBLOCK,    11, 1);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_NSEL,         12, 2);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_MSEL,         16, 8);
LPCLIB_DefineRegBit(CGU_PLL1_CTRL_CLK_SEL,      24, 5);

LPCLIB_DefineRegBit(CCUx_CLK_y_CFG_RUN,         0,  1);
LPCLIB_DefineRegBit(CCUx_CLK_y_CFG_AUTO,        1,  1);
LPCLIB_DefineRegBit(CCUx_CLK_y_CFG_WAKEUP,      2,  1);
LPCLIB_DefineRegBit(CCUx_CLK_y_CFG_DIV,         5,  3);


#define IRC_FREQUENCY                       (12000000ul)

#define PLL1_M_MIN                          (1)
#define PLL1_M_MAX                          (31)
#define PLL1_N_MIN                          (1)
#define PLL1_N_MAX                          (4)
#define PLL1_F_MIN                          (156000ul)
#define PLL1_F_MAX                          (320000ul)
#define CCLK_DIV_MAX                        (31)



/** \addtogroup CLKPWR_Public_Functions
 *  @{
 */


/** Get frequency of an input clock.
 */
static uint32_t __CLKPWR_getInputFrequency (CLKPWR_Source source)
{
    switch (source) {
    case CLKPWR_SOURCE_RTC:
        return 32768ul;
    case CLKPWR_SOURCE_IRC:
        return IRC_FREQUENCY;
    case CLKPWR_SOURCE_ENET_RX:
    case CLKPWR_SOURCE_ENET_TX:
        return LPCLIB_ENET_EXTCLK;
    case CLKPWR_SOURCE_GP_CLKIN:
        return 0;   //TODO
    case CLKPWR_SOURCE_XTAL:
        return LPCLIB_CLKPWR_XTAL_FREQUENCY;

    default:                                            /* We should never come here */
        return 0;
    }
}


static uint32_t __CLKPWR_getPll1Frequency (void)
{
    uint32_t reg = LPC_CGU->PLL1_CTRL;
    CLKPWR_Source sourceClock =
        (CLKPWR_Source)((reg & CGU_PLL1_CTRL_CLK_SEL_Msk) >> CGU_PLL1_CTRL_CLK_SEL_Pos);
    uint32_t inputFrequency = __CLKPWR_getInputFrequency(sourceClock);
    uint32_t p, n, m;

    p = 1u << ((reg & CGU_PLL1_CTRL_PSEL_Msk) >> CGU_PLL1_CTRL_PSEL_Pos);
    if (reg & CGU_PLL1_CTRL_DIRECT_Msk) {
        p = 1;
    }
    n = 1 + ((reg & CGU_PLL1_CTRL_NSEL_Msk) >> CGU_PLL1_CTRL_NSEL_Pos);
    m = 1 + ((reg & CGU_PLL1_CTRL_MSEL_Msk) >> CGU_PLL1_CTRL_MSEL_Pos);

    if (reg & CGU_PLL1_CTRL_BYPASS_Msk) {
        return inputFrequency / p;
    }

    if (reg & CGU_PLL1_CTRL_FBSEL_Msk) {
        return (inputFrequency / n) * m;
    }    

    return ((inputFrequency / n) * m) / p;
}



static uint32_t __CLKPWR_getBaseClock (CLKPWR_BaseClock baseClock) 
{
    CLKPWR_Source outStageSource = (CLKPWR_Source)
        ((LPC_CGU->OUTCLK_CTRL[baseClock] & CGU_OUTCLK_CLK_SEL_Msk) >> CGU_OUTCLK_CLK_SEL_Pos);

    switch (outStageSource) {
    case CLKPWR_SOURCE_RTC:
    case CLKPWR_SOURCE_IRC:
    case CLKPWR_SOURCE_ENET_RX:
    case CLKPWR_SOURCE_ENET_TX:
    case CLKPWR_SOURCE_GP_CLKIN:
    case CLKPWR_SOURCE_XTAL:
        return __CLKPWR_getInputFrequency(outStageSource);

    case CLKPWR_SOURCE_PLL0_AUDIO:
        ;

    case CLKPWR_SOURCE_PLL1:
        return __CLKPWR_getPll1Frequency();

    case CLKPWR_SOURCE_IDIVA:
    case CLKPWR_SOURCE_IDIVB:
    case CLKPWR_SOURCE_IDIVC:
    case CLKPWR_SOURCE_IDIVD:
    case CLKPWR_SOURCE_IDIVE:
        ;

    default:                                            /* We should never come here */
        return 0;
    }
}



/** Return the clock frequency (in Hz) of an internal bus.
 *
 *  \param[in] bus The bus to be queried
 */
uint32_t CLKPWR_getBusClock (CLKPWR_Clock clock)
{
    __IO uint32_t *p;
    uint32_t baseClock;


    /* We need to get the corresponding base clock */
    baseClock = __CLKPWR_getBaseClock((CLKPWR_BaseClock)((clock >> 11) & 0x1F));

    /* Is it enabled or not? */
    if (clock & (1u << 30)) {
        p = (__IO uint32_t *)LPC_CCU2_BASE;
    }
    else {
        p = (__IO uint32_t *)LPC_CCU1_BASE;
    }
    if (p[clock & 0x3FF] & CCUx_CLK_y_CFG_RUN_Msk) {
        if (clock != CLKPWR_CLOCK_EMCDIV) {
            return baseClock;
        }

        /* Special case: EMC_DIV has an optional post-divider by 2. */
        return baseClock / (((p[clock & 0x3FF] & CCUx_CLK_y_CFG_DIV_Msk) >> CCUx_CLK_y_CFG_DIV_Pos) + 1);
    }

    /* Clock is off */
    return 0;
}



/** Enables the crystal oscillator.
 */
static LPCLIB_Result CLKPWR_enableMainOscillator (LPCLIB_Switch activate)
{
    if (activate) {
        if (LPC_CGU->XTAL_OSC_CTRL & CGU_XTAL_OSC_CTRL_ENABLE_Msk) {    /* Already running? */
            if (LPCLIB_CLKPWR_XTAL_FREQUENCY < 16000000ul) {            /* XTAL below 16 MHz? */
                LPC_CGU->XTAL_OSC_CTRL &= ~(CGU_XTAL_OSC_CTRL_HF_Msk);  /* Yes. Use range 1-20 MHz */
            }
            else {
                LPC_CGU->XTAL_OSC_CTRL |= CGU_XTAL_OSC_CTRL_HF_Msk;     /* No. Use range 15-25 MHz */
            }
            LPC_CGU->XTAL_OSC_CTRL &= ~CGU_XTAL_OSC_CTRL_ENABLE_Msk;    /* Start now */

//TODO            while (!(LPC_SC->SCS & SC_SCS_OSCSTAT_Msk))         /* Wait until it's stable */
//                ;
        }
    }
    else {
//TODO do not deactivate if it is current clock source
        LPC_CGU->XTAL_OSC_CTRL |= CGU_XTAL_OSC_CTRL_ENABLE_Msk;         /* Stop */
    }

    return LPCLIB_SUCCESS;
}



/** Attempts to set the CPU clock to the desired frequency.
 *
 *  Sets the CPU clock to the frequency which is closest to the requested
 *  value. Assumes that the PLL input clock has been selected before.
 *
 *  \param[in] targetCpuFrequency Desired CPU frequency (Hz)
 *  \param[out] newCpuFrequency CPU frequency (Hz) after the call
 *  \return ...
 */   //TODO TODO TODO... :-)
LPCLIB_Result CLKPWR_setCpuClock (uint32_t targetCpuFrequency)
{
    uint32_t inFrequency;
    CLKPWR_Source inClock;
    uint32_t m, n, p, div;
    uint32_t mopt, nopt, popt, divopt, enableopt, fbselopt;
    uint32_t f, fpll;
    int32_t delta, maxdelta;


    /* Anything to do at all? */
    if (targetCpuFrequency != SystemCoreClock) {
        /* Connect CPU to IRC. */
        LPC_CGU->OUTCLK_CTRL[4] = (LPC_CGU->OUTCLK_CTRL[4] & ~CGU_OUTCLK_CLK_SEL_Msk)
            | (CLKPWR_SOURCE_IRC << CGU_OUTCLK_CLK_SEL_Pos);

        /* Stop PLL1 */
        LPC_CGU->PLL1_CTRL = CGU_PLL1_CTRL_PD_Msk | CGU_PLL1_CTRL_BYPASS_Msk;

        /* Determine PLL clock input */
#if LPCLIB_CLKPWR_PLL_INPUT_CLOCK == CLKPWR_OSCILLATOR_SYSTEM
        inFrequency = LPCLIB_CLKPWR_XTAL_FREQUENCY;         /* Crystal oscillator */
        inClock = CLKPWR_SOURCE_XTAL;
        CLKPWR_enableMainOscillator(ENABLE);
#else
        inFrequency = IRC_FREQUENCY;                        /* IRC oscillator */
        inClock = CLKPWR_SOURCE_IRC;
#endif
        SystemCoreClock = inFrequency;

        /* Find optimum settings */
        maxdelta = 0x7FFFFFFF;
        enableopt = 0;
        divopt = 1;
        mopt = 6;
        nopt = 1;
        popt = 1;
        fbselopt = 1;

        for (div = 1; div <= 256; div++) {
            f = inFrequency / div;
            delta = f - targetCpuFrequency;
            if (delta < 0) {
                delta = -delta;
            }

            if (delta < maxdelta) {
                maxdelta = delta;
                divopt = div;
                SystemCoreClock = f;
            }
        }

        for (m = PLL1_M_MIN; m <= PLL1_M_MAX; m++) {
            for (n = PLL1_N_MIN; n <= PLL1_N_MAX; n++) {
                for (p = 0; p < 4; p++) {
                    fpll = ((inFrequency / 1000ul) * 2 * m * (1u << p)) / n;
                    if ((fpll >= PLL1_F_MIN) && (fpll <= PLL1_F_MAX)) {
                        f = (inFrequency * m) / n;
                        delta = f - targetCpuFrequency;
                        if (delta < 0) {
                            delta = -delta;
                        }
                        if (delta < maxdelta) {
                            maxdelta = delta;
                            enableopt = 1;
                            mopt = m;
                            nopt = n;
                            popt = p;
                            divopt = div;
                            SystemCoreClock = f;
                        }
    
                        if (f < targetCpuFrequency) {
                            break;                      /* Stop here. Larger dividers won't fit! */
                        }
                    }
                }
            }
        }

        if (enableopt) {
            /* Enable the Main PLL, and wait until it locks. Then connect it. */
            LPC_CGU->PLL1_CTRL =
                (LPC_CGU->PLL1_CTRL & ~(   (CGU_PLL1_CTRL_PD_Msk)
                                      |   (CGU_PLL1_CTRL_BYPASS_Msk)
                                      )
                )
                |   (   (fbselopt << CGU_PLL1_CTRL_FBSEL_Pos)
                    |   (popt << CGU_PLL1_CTRL_PSEL_Pos)
                    |   ((nopt - 1) << CGU_PLL1_CTRL_NSEL_Pos)
                    |   ((mopt - 1) << CGU_PLL1_CTRL_MSEL_Pos)
                    |   (inClock << CGU_PLL1_CTRL_CLK_SEL_Pos)
                    );
                ;

            while (!(LPC_CGU->PLL1_STAT & CGU_PLL1_STAT_LOCK_Msk))
                ;

            LPC_CGU->OUTCLK_CTRL[4] = (LPC_CGU->OUTCLK_CTRL[4] & ~CGU_OUTCLK_CLK_SEL_Msk)
                | (CLKPWR_SOURCE_PLL1 << CGU_OUTCLK_CLK_SEL_Pos);
        }
    }

    return LPCLIB_SUCCESS;
}



/* Sets up the USB clock using the USB PLL. */
LPCLIB_Result CLKPWR_setUsbClock (void)
{
    return LPCLIB_SUCCESS;
}



/* Select an input for a base clock. */
LPCLIB_Result CLKPWR_connectBaseClock (CLKPWR_BaseClock baseClock, CLKPWR_Source sourceClock)
{
    LPC_CGU->OUTCLK_CTRL[(baseClock >> 11) & 0x1F] =
            (LPC_CGU->OUTCLK_CTRL[(baseClock >> 11) & 0x1F] & ~(CGU_OUTCLK_CLK_SEL_Msk))
        |   (sourceClock << CGU_OUTCLK_CLK_SEL_Pos);

    return LPCLIB_SUCCESS;
}



/* TODO: Add definitions for FPU that are missing in CMSIS library V2.1.
 * SCB->CPACR register has been added to the core_cm4.h file!
 */
typedef struct
{
    __IO uint32_t FPCCR;                    /*!< Offset: 0x00  FP Context Control Register                             */
    __IO uint32_t FPCAR;                    /*!< Offset: 0x04  FP Context Address Register                             */
    __IO uint32_t FPDSCR;                   /*!< Offset: 0x08  FP Default Status Control Register                      */
    __I  uint32_t MVFR0;                    /*!< Offset: 0x0C  Media and VFP Feature register 0                        */
    __I  uint32_t MVFR1;                    /*!< Offset: 0x10  Media and VFP Feature Register 1                        */
} FPU_Type;                                                

#define FPU_BASE            (SCS_BASE +  0x0F34)                      /*!< FPU Base Address */
#define FPU                 ((FPU_Type *) FPU_BASE)

/* Enable the FPU. */
void CLKPWR_enableFPU (void)
{
    /* Only try to enable the FPU if it exists! */
    if ((FPU->MVFR0 == 0x10110021) && (FPU->MVFR1 == 0x11000011)) {
        SCB->CPACR |= (0xF << 20);                      /* See ARM M4 TRM 7.3.1 */
    }
}


/** @} */

/** @} CLKPWR */


