/* $Id::                                                               $
 *
 * Copyright (c) 2011, NXP Semiconductors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "lpc43xx_libconfig.h"


/** \addtogroup GPIO
 *  @{
 */

#include "lpc43xx_gpio.h"
#include "lpc43xx_clkpwr.h"
#include "lpclib_osal.h"


/** GPIO module context. */
static struct GPIO_Context {
    LPCLIB_DECLARE_MUTEX(mutex);
#if LPCLIB_GPIO_INTERRUPTS
    LPCLIB_Callback callbacks[__NUM_GPIO__];
#endif
} gpioContext;




void GPIO_open (void)
{
//TODO    CLKPWR_enableClock(CLKPWR_CLOCKSWITCH_GPIO);
    LPCLIB_OSALMutexCreate((LPCLIB_OSAL_NAME *)"", &gpioContext.mutex);
#if LPCLIB_RTOS == 0
    gpioContext.mutex = &gpioContext.mutex_value;
#endif
}


void GPIO_setDir32 (GPIO_Port32 port, uint32_t value, uint32_t mask)
{
    if (LPCLIB_OSALSemaphoreAcquire(gpioContext.mutex, LPCLIB_OSAL_TIMEOUT_INFINITE, LPCLIB_OSAL_CALL_DEFAULT) == LPCLIB_OSAL_ERROR_NONE) {
        LPC_GPIO_PORT->DIR[port] = (LPC_GPIO_PORT->DIR[port] & ~(~value & mask)) | (value & mask);

        LPCLIB_OSALSemaphoreRelease(gpioContext.mutex, 0, LPCLIB_OSAL_CALL_DEFAULT);
    }
}

void GPIO_setDirBit (GPIO_Pin pin, LPCLIB_Switch outputEnable)
{
    if (LPCLIB_OSALSemaphoreAcquire(gpioContext.mutex, LPCLIB_OSAL_TIMEOUT_INFINITE, LPCLIB_OSAL_CALL_DEFAULT) == LPCLIB_OSAL_ERROR_NONE) {
        LPC_GPIO_PORT->DIR[pin / 32] = (LPC_GPIO_PORT->DIR[pin / 32] & ~(1u << (pin % 32))) | (outputEnable << (pin % 32));

        LPCLIB_OSALSemaphoreRelease(gpioContext.mutex, 0, LPCLIB_OSAL_CALL_DEFAULT);
    }
}


/* Set GPIO options. */
void GPIO_ioctl (const GPIO_Config *pConfig)
{
#if LPCLIB_GPIO_INTERRUPTS
    int port;
    int intNum;
    uint32_t mask;
#endif


    switch (pConfig->opcode) {
#if LPCLIB_GPIO_INTERRUPTS
    case GPIO_OPCODE_SET_CALLBACK:
        port = (pConfig->interrupt.pin >> 0) & 3;
        gpioContext.callbacks[port] = pConfig->callbackFromISR;
        break;

    case GPIO_OPCODE_CONFIGURE_INTERRUPT:
        intNum = pConfig->interrupt.interruptLine;
        mask = 1u << intNum;

        LPC_GPIO_PIN_INT->CIENR = mask;                 /* Always mask interrupts during configuration */
        LPC_GPIO_PIN_INT->CIENF = mask;
        NVIC_DisableIRQ((IRQn_Type)(PIN_INT0_IRQn + intNum));

        if (pConfig->interrupt.enable) {
            switch (pConfig->interrupt.mode) {
            case GPIO_INT_FALLING_EDGE:
                LPC_GPIO_PIN_INT->ISEL &= ~mask;        /* Edge sensitive */
                LPC_GPIO_PIN_INT->SIENF = mask;         /* Falling edge */
                break;
            case GPIO_INT_RISING_EDGE:
                LPC_GPIO_PIN_INT->ISEL &= ~mask;        /* Edge sensitive */
                LPC_GPIO_PIN_INT->SIENR = mask;         /* Rising edge */
                break;
            case GPIO_INT_BOTH_EDGES:
                LPC_GPIO_PIN_INT->ISEL &= ~mask;        /* Edge sensitive */
                LPC_GPIO_PIN_INT->SIENR = mask;         /* Rising edge */
                LPC_GPIO_PIN_INT->SIENF = mask;         /* Falling edge */
                break;
            case GPIO_INT_LOW_LEVEL:
                LPC_GPIO_PIN_INT->ISEL |= mask;         /* Level sensitive */
                LPC_GPIO_PIN_INT->SIENR = mask;         /* Low level */
                break;
            case GPIO_INT_HIGH_LEVEL:
                LPC_GPIO_PIN_INT->ISEL |= mask;         /* Level sensitive */
                LPC_GPIO_PIN_INT->SIENF = mask;         /* High level */
                break;
            }

            /* Connect selected GPIO to the indiviual interrupt line */
            LPC_SCU->PINTSEL[intNum / 4] =
                    (LPC_SCU->PINTSEL[intNum / 4] & ~(0xFF << (intNum % 4)))
                |   ((pConfig->interrupt.pin / 32) << 5)
                |   ((pConfig->interrupt.pin % 32) << 0);
    
            NVIC_EnableIRQ((IRQn_Type)(PIN_INT0_IRQn + intNum));
        }
        break;
#endif

    default:
        break;
    }
}



#if LPCLIB_GPIO_INTERRUPTS

static void GPIO_CommonIRQHandler (GPIO_Port32 port)
{
#if 0
    LPCLIB_Event event;

    uint32_t status = __gpio[port]->MIS;                /* Read interrupt sources */
    __gpio[port]->IC = status;                          /* Clear the requests */

    if (gpioContext.callbacks[port]) {
        event.id = LPCLIB_EVENTID_GPIO;
        event.block = port;
        event.parameter = (void *)status;
        gpioContext.callbacks[port](event);
    }
#endif
}


#endif



/** @} */

