/* $Id::                                                               $
 *
 * Copyright (c) 2011, NXP Semiconductors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "lpc43xx_libconfig.h"

#if LPCLIB_LCD

/** \file
 *  \brief LCD driver implementation.
 *
 *  This file contains the driver code for the LCD peripheral.
 *
 *  \author NXP Semiconductors
 */




#include "lpc43xx_lcd.h"
#include "lpc43xx_clkpwr.h"
#include "lpclib_osal.h"


/** (Selected) Field definitions for hardware register LCD_CTRL. */
LPCLIB_DefineRegBit(LCD_CTRL_LCDEN,             0,  1);
LPCLIB_DefineRegBit(LCD_CTRL_LCDPWR,            11, 1);


/** Local context of LCD. */
static struct LCD_Context {
    LCD_Name name;                                      /**< LCD identifier */
} lcdContext;



/* Open the LCD interface. */
LPCLIB_Result LCD_open (LCD_Name name, LCD_Handle *pHandle)
{
//    CLKPWR_deassertPeripheralReset(i2cReset[bus]);      /* Release I2C reset */
    CLKPWR_enableClock(CLKPWR_CLOCKSWITCH_LCD);         /* Enable I2C peripheral clock */

    /* Cannot open an LCD twice */
if(1){//    if (!(i2c->CONSET & I2C_CONSET_I2EN)) {
        *pHandle = &lcdContext;                         /* Return handle */
        NVIC_EnableIRQ(LCD_IRQn);                       /* Enable interrupts */

        return LPCLIB_SUCCESS;
    }

    *pHandle = LPCLIB_INVALID_HANDLE;

    return LPCLIB_BUSY;

}



/* Close the LCD interface. */
void LCD_close (LCD_Handle *pHandle)
{
    if (*pHandle == LPCLIB_INVALID_HANDLE) {
        return;
    }

    NVIC_DisableIRQ(LCD_IRQn);                          /* Disable interrupts */
    CLKPWR_disableClock(CLKPWR_CLOCKSWITCH_LCD);        /* Disable LCD peripheral clock */

    *pHandle = LPCLIB_INVALID_HANDLE;
}



/* Configure the LCD block. */
void LCD_ioctl (LCD_Handle handle, const LCD_Config *pConfig)
{
    uint32_t divider;
    int bypass;
    int extClock;

    if (handle == LPCLIB_INVALID_HANDLE) {
        return;
    }

    switch (pConfig->opcode) {
    case LCD_OPCODE_SET_TIMINGS:
        LPC_LCD->TIMH =
            ((pConfig->timings.sizeX / 16 - 1)  << 2)   |
            ((pConfig->timings.syncWidthH - 1)  << 8)   |
            ((pConfig->timings.frontPorchH - 1) << 16)  |
            ((pConfig->timings.backPorchH - 1)  << 24);

        LPC_LCD->TIMV =
            ((pConfig->timings.sizeY - 1)       << 0)   |
            ((pConfig->timings.syncWidthV - 1)  << 10)  |
            (pConfig->timings.frontPorchV       << 16)  |
            (pConfig->timings.backPorchV        << 24);

        CLKPWR_connectBaseClock(CLKPWR_BASECLOCK_LCD, CLKPWR_SOURCE_PLL1); //TODO
        divider = 12-2;            //TODO
        bypass = 0;                //TODO
        extClock = pConfig->timings.externalClockFrequency ? 1 : 0;
        LPC_LCD->POL =
            ((divider & 0x1F)                   << 0)   |
            (extClock                           << 5)   |
            (0                                  << 6)   |
            (pConfig->timings.polarityVSYNC     << 11)  |
            (pConfig->timings.polarityHSYNC     << 12)  |
            (pConfig->timings.polarityClock     << 13)  |
            (pConfig->timings.polarityENAB      << 14)  |
            ((pConfig->timings.sizeX - 1)       << 16)  |
            (bypass                             << 26)  |
            ((divider >> 5)                     << 27);

        LPC_LCD->CTRL =
            (0 << 0)                                    |
            (pConfig->timings.bitsPerPixel << 1)        |
            (1u << 5)                                   |   /* TFT */
            (0 << 8)                                    |   /* RGB/BGR */
            (0 << 9)                                    |   /* Little endian byte order */
            (0 << 10)                                   |   /* Little endian pixel order */
            (0 << 11)                                   |   /* Power off */
            0;
        break;

    case LCD_OPCODE_SET_POWER:
        if (pConfig->powered) {
            LPC_LCD->CTRL |= LCD_CTRL_LCDEN_Msk | LCD_CTRL_LCDPWR_Msk;
        }
        else {
            LPC_LCD->CTRL &= ~(LCD_CTRL_LCDEN_Msk | LCD_CTRL_LCDPWR_Msk);
        }
        break;
    }
}


/** LCD interrupt handler.
 */
void LCD_IRQHandler (void)
{
}

/** @} */

/** @} addtogroup LCD */

#endif  /* #ifdef LPCLIB_LCD */

