/* $Id::                                                               $
 *
 * Copyright (c) 2011, NXP Semiconductors
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 * Neither the name of the author nor the names of its contributors may be used
 * to endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "lpc43xx_libconfig.h"

#if LPCLIB_EMC

/** \file
 *  \brief EMC driver implementation.
 *
 *  This file contains the driver code for the EMC peripheral.
 *
 *  \author NXP Semiconductors
 */


#include "lpc43xx_emc.h"
#include "lpc43xx_clkpwr.h"
#include "lpclib_osal.h"


/** Field definition for hardware register EMCControl. */
enum {
    EMC_CONTROL_ENABLE = (1u << 0),
    EMC_CONTROL_ADDRESS_MIRROR = (1u << 1),
    EMC_CONTROL_LOW_POWER = (1u << 2),
};

/** Field definition for hardware register EMCDynamicControl. */
enum {
    EMC_DYNCONTROL_CE_NORMAL = (0u << 0),
    EMC_DYNCONTROL_CE_CONTINUOUS = (1u << 0),
    EMC_DYNCONTROL_CS_NORMAL = (0u << 1),
    EMC_DYNCONTROL_CS_CONTINUOUS = (1u << 1),
    EMC_DYNCONTROL_SR_DISABLE = (0u << 2),
    EMC_DYNCONTROL_SR_ENABLE = (1u << 2),
    EMC_DYNCONTROL_MMC_ENABLE = (0u << 5),
    EMC_DYNCONTROL_MMC_DISABLE = (1u << 5),
    EMC_DYNCONTROL_I_NORMAL = (0u << 7),
    EMC_DYNCONTROL_I_MODE = (1u << 7),
    EMC_DYNCONTROL_I_PALL = (2u << 7),
    EMC_DYNCONTROL_I_NOP = (3u << 7),
    EMC_DYNCONTROL_DP_NORMAL = (0u << 13),
    EMC_DYNCONTROL_DP_DEEPSLEEP = (1u << 13),
};

/** Field definition for hardware register EMCDynamicReadConfig. */
enum {
    EMC_DYNREADCONFIG_CLOCKOUT_DELAYED = 0,
    EMC_DYNREADCONFIG_COMMAND_DELAYED = 1,
    EMC_DYNREADCONFIG_COMMAND_DELAYED_PLUS1 = 2,
    EMC_DYNREADCONFIG_COMMAND_DELAYED_PLUS2 = 3,
};


/* Base address of dynamic memory bank n */
const uint32_t emcDynamicBaseAddress[4] = {
    0x28000000, 0x30000000, 0x60000000, 0x70000000
};

#define EMC_SDRAM_BASE_ADDRESS(n)   emcDynamicBaseAddress[n]


/** Local device context. */
static struct EMC_Context {
    _Bool halfRate;                         /**< EMC runs on half the CCLK frequency */
} emcContext;



/* Open the EMC. */
LPCLIB_Result EMC_open (EMC_Handle *pHandle)
{
    uint32_t emcClock;

    if (*pHandle) {
        return LPCLIB_BUSY;
    }

//TODO    CLKPWR_deassertPeripheralReset(CLKPWR_RESET_EMC);   /* Release EMC reset */
    CLKPWR_enableClock(CLKPWR_CLOCKSWITCH_EMC);         /* Enable EMC clock */

    /* Check whether we can use full CCLK speed */
    emcContext.halfRate = false;
    emcClock = CLKPWR_getBusClock(CLKPWR_CLOCK_CPU);
    if (emcClock > 120000000ul) {
        emcContext.halfRate = true;
        emcClock /= 2;
    }

    LPC_EMC->CONTROL = EMC_CONTROL_ENABLE;              /* Disable address mirror, enable EMC */
    
    LPC_SCU->EMCDELAYCLK = 0x00004444;

    /* Auto select SDRAM read strategy, depending on bus clock frequency */
    if (emcClock <= 60000000ul) {
        LPC_EMC->DYNAMICREADCONFIG = EMC_DYNREADCONFIG_CLOCKOUT_DELAYED;
    }
    else {
        LPC_EMC->DYNAMICREADCONFIG = EMC_DYNREADCONFIG_COMMAND_DELAYED;
    }
    *pHandle = &emcContext;                             /* Return handle */

    return LPCLIB_SUCCESS;
}



#if LPCLIB_EMC_CLOSE
/* Close the EMC. */
LPCLIB_Result EMC_close (EMC_Handle *pHandle)
{
    if (*pHandle == LPCLIB_INVALID_HANDLE) {
        return LPCLIB_ILLEGAL_PARAMETER;
    }

    LPC_EMC->Control = 0;                               /* Disable the EMC */
    CLKPWR_disableClock(CLKPWR_CLOCKSWITCH_EMC);        /* Disable the clock */

    *pHandle = LPCLIB_INVALID_HANDLE;

    return LPCLIB_SUCCESS;
}
#endif


/* Configure the EMC block. */
void EMC_ioctl (EMC_Handle handle, const EMC_Config *pConfig)
{
    if (handle == LPCLIB_INVALID_HANDLE) {
        return;
    }

    LPC_EMC_Type * const emc = LPC_EMC;
    int n;
    int i;
    uint32_t temp;
    int modeReg;
    int offset;


    while (1) {
        switch (pConfig->opcode) {
        case EMC_OPCODE_DYNAMIC_GLOBALS:                    /* Dynamic memory timings (all devices) */
            emc->DYNAMICREFRESH = pConfig->dynTimings.refreshCycles;
            emc->DYNAMICRP = pConfig->dynTimings.trp;
            emc->DYNAMICRAS = pConfig->dynTimings.tras;
            emc->DYNAMICSREX = pConfig->dynTimings.tsrex;
            emc->DYNAMICAPR = pConfig->dynTimings.tapr;
            emc->DYNAMICDAL = pConfig->dynTimings.tdal;
            emc->DYNAMICWR = pConfig->dynTimings.twr;
            emc->DYNAMICRC = pConfig->dynTimings.trc;
            emc->DYNAMICRFC = pConfig->dynTimings.trfc;
            emc->DYNAMICXSR = pConfig->dynTimings.txsr;
            emc->DYNAMICRRD = pConfig->dynTimings.trrd;
            emc->DYNAMICMRD = pConfig->dynTimings.tmrd;
            break;

        case EMC_OPCODE_DYNAMIC_DEVICE:                     /* Configure an SDRAM device */
            n = (int)(pConfig->dynDevice.device);
            emc->D[n].DYNAMICRASCAS = (pConfig->dynDevice.ras << 0) |
                                      (pConfig->dynDevice.cas << 8);
            emc->D[n].DYNAMICCONFIG =
                (pConfig->dynDevice.busWidth << 14) |
                (pConfig->dynDevice.lowPower << 12) |
                (pConfig->dynDevice.org << 7) |
                (pConfig->dynDevice.lowPower << 3);
            emc->DYNAMICCONTROL =
                EMC_DYNCONTROL_I_NOP |
                EMC_DYNCONTROL_CS_CONTINUOUS |
                EMC_DYNCONTROL_CE_CONTINUOUS;
            for (i = 0; i < 10000; i++);     //TODO (min. 100us)

            emc->DYNAMICCONTROL =
                EMC_DYNCONTROL_I_PALL |
                EMC_DYNCONTROL_CS_CONTINUOUS |
                EMC_DYNCONTROL_CE_CONTINUOUS;
            temp = emc->DYNAMICREFRESH;
            emc->DYNAMICREFRESH = 4;
            for (i = 0; i < 10000; i++);       //TODO
            emc->DYNAMICREFRESH = temp;

            emc->DYNAMICCONTROL =
                EMC_DYNCONTROL_I_MODE |
                EMC_DYNCONTROL_CS_CONTINUOUS |
                EMC_DYNCONTROL_CE_CONTINUOUS;

            /* Determine content of SDRAM's mode register.
             * Default: 32-bit bus: 4 read cycles/burst.
             * Use same CAS delay as programmed in EMC.
             */
            modeReg = 0x02 | ((pConfig->dynDevice.cas - 0) << 4);
            if (pConfig->dynDevice.busWidth == EMC_DYNAMIC_BUS_16) {
                modeReg |= 1;                       /* 16-bit bus: 8 read cycles/burst */
            }

            /* Determine the dummy read address for mode register programming. */
            offset = 1 + 8 + ((pConfig->dynDevice.org & 1) ^ 1) + ((pConfig->dynDevice.org >> 2) & 7) / 2; //TODO
            if (pConfig->dynDevice.busWidth == EMC_DYNAMIC_BUS_32) {
                ++offset;                           /* 32-bit bus: one extra shift */
            }
            if (!pConfig->dynDevice.lowPower) {
                ++offset;                           /* Low power mode, 2 banks: one extra shift */
                if (pConfig->dynDevice.org & (7u << 2)) {
                    ++offset;                       /* Another shift for 4 banks */
                }
            }
            i = *((volatile uint32_t *)(EMC_SDRAM_BASE_ADDRESS(n) + (modeReg << offset)));  /* Set mode register! */

            emc->DYNAMICCONTROL =
                EMC_DYNCONTROL_I_NORMAL |
                EMC_DYNCONTROL_CS_NORMAL |
                EMC_DYNCONTROL_CE_NORMAL;
            emc->D[n].DYNAMICCONFIG |= (1u << 19);      /* Enable buffering */
            break;

        case EMC_OPCODE_STATIC_DEVICE:
            emc->S[pConfig->staticDevice.device].STATICCONFIG =
                pConfig->staticDevice.busWidth |
                pConfig->staticDevice.multi;

            temp = pConfig->staticDevice.writePulseDelay;
            if (temp == 0) {
                temp = 1;
            }
            emc->S[pConfig->staticDevice.device].STATICWAITWEN = temp - 1;

            temp = temp + pConfig->staticDevice.writePulseWidth;
            emc->S[pConfig->staticDevice.device].STATICWAITWR = temp - 2;

            temp = pConfig->staticDevice.readCycleWidth;
            if (temp == 0) {
                temp = 1;
            }
            emc->S[pConfig->staticDevice.device].STATICWAITRD = temp - 1;
            temp = pConfig->staticDevice.readCycleWidthPage;
            if (temp == 0) {
                temp = 1;
            }
            emc->S[pConfig->staticDevice.device].STATICWAITPAGE = temp - 1;
            emc->S[pConfig->staticDevice.device].STATICWAITOEN =
                pConfig->staticDevice.readOeDelay;
            break;
        }

        if (!pConfig->continued) {
            break;
        }

        ++pConfig;                                      /* More config's follow */
    }
}


/** @} */

/** @} addtogroup EMC */

#endif  /* #ifdef LPCLIB_EMC */

