/**
 ******************************************************************************
 * @file	OSAL_FreeRTOS.c
 * @brief	OS wrapper for FreeRTOS.
 * @version	1.0
 * @date	02.06.2010
 * @author	nxp16889
 *
 * @details This module contain code for using FreeRTOS as the underneath OS.
 * 			At least all functions here must be implemented when using a certain
 * 			OS with OSAL.
 *
 * ----------------------------------------------------------------------------\n
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * NXP Semiconductors assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. NXP Semiconductors
 * reserves the right to make changes in the software without
 * notification. NXP Semiconductors also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 * ----------------------------------------------------------------------------\n
 ******************************************************************************
 */

#include "OSAL_FreeRTOS.h"
#include "FreeRTOSConfig.h"

#if 0
#ifndef portEND_SWITCHING_ISR
    #define portEND_SWITCHING_ISR(taskSwitch)   \
        if (taskSwitch) {                       \
            portYIELD_FROM_ISR();               \
        }
#endif
#endif


void OSW_EnterCritical (void)
{
    portENTER_CRITICAL();
}
void OSW_ExitCritical (void)
{
	portEXIT_CRITICAL();
}
void OSW_EnterISR(OSAL_ARG *pHighPrioTaskSwitch)
{
	*pHighPrioTaskSwitch = OSAL_FALSE;
}
void OSW_ExitISR(OSAL_ARG *pHighPrioTaskSwitch)
{
	//this is a macro, which need access to the variable, not
	//the pointer, so pass the content where the pointer points to.
	portEND_SWITCHING_ISR(*pHighPrioTaskSwitch);
}
void OSW_DisableInterrupts (void)
{
	portDISABLE_INTERRUPTS();
}
void OSW_EnableInterrupts (void)
{
	portENABLE_INTERRUPTS();
}

OSAL_ERROR OSW_Start (void)
{
	vTaskStartScheduler();

	return OSAL_ERROR_NONE;
}

OSAL_TICK OSW_GetTickCount (void)
{
	return ((OSAL_TICK)xTaskGetTickCount());
}

/*
 ****************************************************************************
 * author nxp16889
 *
 * Implement task handling here.
 ****************************************************************************
 */

OSAL_ERROR OSW_TaskCreate(OSAL_NAME *pName,
                          OSAL_TASK_FUNCTION func,
                          OSAL_STACK_SIZE stackSize,
                          OSAL_TASK_ARGS *pArgs,
                          OSAL_PRIO prio,
                          OSAL_TASK *pHandle)
{
signed portBASE_TYPE retValue;
xTaskHandle task;

    //be sure that pHandle is NULL in case of an error. Application code than can also check for NULL for handle.
    if (pHandle) {
        *pHandle = NULL;
    }

	retValue = xTaskCreate(	func,
							(const signed char*)pName,
							(unsigned short)stackSize,
							pArgs,
							(unsigned portBASE_TYPE)prio,
							&task);

	switch(retValue)
    {
        case pdPASS:
            if (pHandle) {
                *pHandle = task;
            }
	    	return OSAL_ERROR_NONE;

	    default:
	    	return OSAL_ERROR_INTERNAL_ERROR;
    }
}

OSAL_ERROR OSW_TaskDelete(OSAL_TASK handle)
{
	vTaskDelete(handle);

	return OSAL_ERROR_NONE;
}

OSAL_ERROR	OSW_TaskSuspend(OSAL_TASK handle)
{
#if INCLUDE_vTaskSuspend
    vTaskSuspend(handle);
#else
    (void) handle;
#endif

	return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TaskResume(OSAL_TASK handle)
{
#if INCLUDE_vTaskSuspend
	vTaskResume(handle);
#else
    (void) handle;
#endif

	return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TaskPrioSet(OSAL_TASK handle, OSAL_PRIO prio)
{
#if INCLUDE_vTaskPrioritySet
    vTaskPrioritySet(handle, prio);
#else
    (void) handle;
    (void) prio;
#endif

	return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TaskPrioGet(OSAL_TASK handle, OSAL_PRIO *pPrio)
{
#if INCLUDE_vTaskPrioritySet
	*pPrio = (OSAL_PRIO)uxTaskPriorityGet(handle);

	return OSAL_ERROR_NONE;
#else
    (void) handle;
    (void) pPrio;

    return OSAL_ERROR_NOT_SUPPORTED;
#endif
}


OSAL_ERROR OSW_TaskDelay(OSAL_TASK handle, OSAL_TICK *pStartTime, OSAL_TICK ticks)
{
#if INCLUDE_vTaskDelayUntil || INCLUDE_vTaskDelay
    (void) handle;


    OSAL_TICK timeDelay;

	timeDelay = ticks/(OSAL_TICK)portTICK_RATE_MS;

	//when start time is 0 (not specified), we just call vTaskDelay, which is not accurate.
	if(pStartTime == NULL)
	{
		vTaskDelay(timeDelay);
	}
	//we call vTaskDelayUntil() which is accurate by considering the given start time,
	//normally the last
	else
	{
		vTaskDelayUntil(pStartTime, timeDelay);
	}

	return OSAL_ERROR_NONE;
#else
    (void) handle;
    (void) pStartTime;
    (void) ticks;

    return OSAL_ERROR_NOT_SUPPORTED;
#endif
}

/*
 ****************************************************************************
 * author nxp16889
 *
 * Implement queue handling here.
 ****************************************************************************
 */


OSAL_ERROR OSW_QueueCreate(OSAL_NAME *pName,
                           OSAL_COUNT queueSize,
                           OSAL_MESSAGE_SIZE itemSize,
                           OSAL_QUEUE *pHandle )
{
    (void) pName;


    xQueueHandle queue;

	*pHandle = NULL;

	queue = xQueueCreate(queueSize, itemSize);

	if(queue == NULL)
	{
		return OSAL_ERROR_OUT_OF_MEMORY;
	}

	*pHandle = queue;

	return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_QueueDelete(OSAL_QUEUE handle)
{
	vQueueDelete(handle);

	return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_QueuePost(OSAL_QUEUE handle,
                         OSAL_MESSAGE *pMsg,
                         OSAL_QUEUE_LOCATION location,
                         OSAL_TIMEOUT timeout,
                         OSAL_ARG *pHighPrioTaskSwitch,
                         OSAL_CALL_OPTION callFromISR)
{
portBASE_TYPE rc;

	if(callFromISR)
	{
		switch(location)
		{
			case OSAL_QUEUE_BACK:
				rc = xQueueSendToBackFromISR(handle, pMsg, pHighPrioTaskSwitch);
				break;
			case OSAL_QUEUE_FRONT:
				rc = xQueueSendToFrontFromISR(handle, pMsg, pHighPrioTaskSwitch);
				break;
			default:
				rc = xQueueSendToBackFromISR(handle, pMsg, pHighPrioTaskSwitch);
				break;
		}
	}
	else
	{
		switch(location)
		{
			case OSAL_QUEUE_BACK:
				rc = xQueueSendToBack(handle, pMsg, timeout);
				break;
			case OSAL_QUEUE_FRONT:
				rc = xQueueSendToFront(handle, pMsg, timeout);
				break;
			default:
				rc = xQueueSendToBack(handle, pMsg, timeout);
				break;
		}
	}

	if(rc == pdPASS)
	{
		return OSAL_ERROR_NONE;
	}
	else
	{
		return OSAL_ERROR_QUEUE_FULL;
	}
}

/**
 **********************************************************************
 * @fn			OSAL_ERROR OSW_QueuePend(OSAL_QUEUE handle,
 * 											OSAL_MESSAGE *pMsg,
 * 											OSAL_TIMEOUT timeout,
 * 											OSAL_ARG *pHighPrioTaskSwitch,
 * 											OSAL_CALL_OPTION callFromISR)
 *
 * @brief		Enter brief description for this method.
 *
 * @param[in]	handle 					Handle to the queue.
 * @param[out]	*pMsg 					Pointer to store the received message.
 * @param[in]	timeout 				Timeout to wait.
 * @param[out]	*pHighPrioTaskSwitch 	Pointer to the HighPrioTaskSwitch variable used when calling from ISR.
 * @param[in]	callFromISR 			Indicates whether or not this function is called from ISR.
 *
 * @return 		OSAL_ERROR
 *********************************************************************
 */
OSAL_ERROR OSW_QueuePend(OSAL_QUEUE handle,
                            OSAL_MESSAGE *pMsg,
                            OSAL_TIMEOUT timeout,
                            OSAL_ARG *pHighPrioTaskSwitch,
                            OSAL_CALL_OPTION callFromISR)
{
portBASE_TYPE rc;

	if(callFromISR)
	{
		rc = xQueueReceiveFromISR(handle, pMsg, pHighPrioTaskSwitch);
	}
	else
	{
		rc = xQueueReceive(handle, pMsg, timeout);
	}

	if(rc == pdPASS)
	{
		return OSAL_ERROR_NONE;
	}
	else
	{
		return OSAL_ERROR_QUEUE_EMPTY;
	}
}

OSAL_ERROR OSW_QueuePeek(OSAL_QUEUE handle, OSAL_MESSAGE *pMsg, OSAL_TIMEOUT timeout)
{
portBASE_TYPE rc;

	rc = xQueuePeek(handle, pMsg, timeout);

	if(rc == pdPASS)
	{
		return OSAL_ERROR_NONE;
	}
	else
	{
		return OSAL_ERROR_QUEUE_EMPTY;
	}
}
OSAL_ERROR OSW_QueueCount(OSAL_QUEUE handle, OSAL_COUNT *pCount)
{
	*pCount = uxQueueMessagesWaiting(handle);

	return OSAL_ERROR_NONE;
}

/*
 ****************************************************************************
 *
 * Implement timer handling here.
 ****************************************************************************
 */

OSAL_ERROR OSW_TimerCreate(OSAL_NAME *pName,
                           OSAL_TIMER_TYPE type,
                           PHANDLE magic,
                           U32 initValue,
                           U32 reloadValue,
                           OSAL_CALLBACK callback,
                           OSAL_TIMER *pHandle)
{
#if configUSE_TIMERS
    (void) initValue;

    xTimerHandle timer;

    timer = xTimerCreate((const signed char *)pName,
                         reloadValue,
                         (type == OSAL_TIMER_PERIODIC) ? pdTRUE : pdFALSE,
                         magic,
                         callback);

    *pHandle = timer;

    if (!timer) {
        return OSAL_ERROR_OUT_OF_MEMORY;
    }
#else
    (void) pName;
    (void) type;
    (void) magic;
    (void) initValue;
    (void) reloadValue;
    (void) callback;
    (void) pHandle;
#endif

    return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TimerDelete(OSAL_TIMER handle)
{
#if configUSE_TIMERS
    if (xTimerDelete(handle, OSAL_TIMEOUT_NONE) != pdPASS) {
        return OSAL_ERROR_QUEUE_FULL;
    }
#else
    (void) handle;
#endif

    return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TimerStart(OSAL_TIMER handle)
{
#if configUSE_TIMERS
    if (xTimerStart(handle, OSAL_TIMEOUT_NONE) != pdPASS) {
        return OSAL_ERROR_QUEUE_FULL;
    }
#else
    (void) handle;
#endif

    return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TimerStop(OSAL_TIMER handle)
{
#if configUSE_TIMERS
    if (xTimerStop(handle, OSAL_TIMEOUT_NONE) != pdPASS) {
        return OSAL_ERROR_QUEUE_FULL;
    }
#else
    (void) handle;
#endif

    return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TimerGetTime(OSAL_TIMER handle, U32 *pCount)
{
    //TODO
    (void) handle;
    (void) pCount;

    return OSAL_ERROR_NONE;
}

OSAL_ERROR OSW_TimerSetTime(OSAL_TIMER handle, U32 initValue, U32 reloadValue)
{
    //TODO
    (void) handle;
    (void) initValue;
    (void) reloadValue;

    return OSAL_ERROR_NONE;
}


OSAL_ERROR OSW_TimerGetMagic(OSAL_TIMER handle, PHANDLE *pMagic)
{
    *pMagic = pvTimerGetTimerID(handle);

    return OSAL_ERROR_NONE;
}




OSAL_ERROR OSW_MemAlloc(OSAL_SIZE size, OSAL_MEMORY *pHandle)
{
    (*pHandle) = pvPortMalloc(size);

    if (*pHandle == NULL) {
        return OSAL_ERROR_OUT_OF_MEMORY;
    }

    return OSAL_ERROR_NONE;
}


OSAL_ERROR OSW_MemRelease(OSAL_MEMORY handle)
{
    if (!handle) {
        return OSAL_ERROR_INVALID_PARAMETER;
    }

    vPortFree(handle);

    return OSAL_ERROR_NONE;
}





/*
 ****************************************************************************
 * author nxp16889
 *
 * Implement semaphore handling here.
 ****************************************************************************
 */
OSAL_ERROR OSW_SemaphoreCreate(OSAL_NAME *pName,
                               OSAL_COUNT maxCount,
                               OSAL_COUNT initCount,
                               OSAL_SEMAPHORE *pHandle)
{
    (void) pName;


    xSemaphoreHandle sem;

	*pHandle = NULL;

	/*
     ****************************************************************************
     * author nxp16889
     *
     * Check for parameters - maxCount must be greater than 0 and initCount must
     * be equal or lower than maxCount
     ****************************************************************************
     */
	if(maxCount == 0 || initCount > maxCount) return OSAL_ERROR_INVALID_PARAMETER;

	if(maxCount == 1)
	{
		/*
         ****************************************************************************
         * author nxp16889
         *
         * If maxCount is set to one, we interpret this as an binary semaphore.
         ****************************************************************************
         */
		vSemaphoreCreateBinary(sem);
		//if no semaphore could be created....
		if(sem == NULL)
		{
			return OSAL_ERROR_OUT_OF_MEMORY;
		}
	}
	else
	{
		/*
         ****************************************************************************
         * author nxp16889
         *
         * This is a counting semaphore. Lets create this here...
         ****************************************************************************
         */
#if configUSE_COUNTING_SEMAPHORES
        sem = xSemaphoreCreateCounting(maxCount, initCount);

		if(sem == NULL)
		{
			return OSAL_ERROR_OUT_OF_MEMORY;
		}
#else
        return OSAL_ERROR_NOT_SUPPORTED;
#endif
    }

	//save the semaphore handle
	*pHandle = sem;

	return OSAL_ERROR_NONE;
}

void OSW_SemaphoreDelete(OSAL_SEMAPHORE handle)
{
    vQueueDelete(handle);
}

OSAL_ERROR OSW_SemaphoreAcquire(OSAL_SEMAPHORE handle, OSAL_TIMEOUT timeout, OSAL_CALL_OPTION callFromISR)
{
    (void) callFromISR;     //TODO calling from ISR is rare. Should we support it?

	if(xSemaphoreTake(handle, timeout) == pdPASS)
	{
		return OSAL_ERROR_NONE;
	}

	return OSAL_ERROR_FAILED;
}

OSAL_ERROR OSW_SemaphoreRelease(OSAL_SEMAPHORE handle, OSAL_ARG *pHighPrioTaskSwitch, OSAL_CALL_OPTION callFromISR)
{

	if(callFromISR)
	{
		if(xSemaphoreGiveFromISR(handle, pHighPrioTaskSwitch) != pdTRUE)
		{
			return OSAL_ERROR_QUEUE_EMPTY;
		}
	}
	else
	{
		if(xSemaphoreGive(handle) != pdPASS)
		{
			return OSAL_ERROR_FAILED;
		}
	}

	return OSAL_ERROR_NONE;
}

/*
 ****************************************************************************
 * author nxp16889
 *
 * Implement mutex handling here.
 ****************************************************************************
 */

OSAL_ERROR OSW_MutexCreate(OSAL_NAME *pName, OSAL_MUTEX *pHandle)
{
    (void) pName;


    xSemaphoreHandle sem;

	sem = xSemaphoreCreateMutex();

	if(sem == NULL)
	{
		return OSAL_ERROR_OUT_OF_MEMORY;
	}

	*pHandle = sem;

	return OSAL_ERROR_NONE;
}

void OSW_MutexDelete(OSAL_MUTEX handle)
{
    vQueueDelete(handle);
}


/**
 **********************************************************************
 * @fn			void vApplicationIdleHook(void)
 *
 * @brief		Idle Hook function for FreeRTOS. This function will be called
 * 				when no other task is in active state. Use this function as
 * 				a background worker to do thing with low priority. Note that
 * 				the function name is fixed and cannot be changed without
 * 				modifying the FreeRTOS source code. Therefore, we just call
 * 				the @ref OSALIdleHook() function which has been declared as
 * 				WEAK and thus can be overwritten by the application layer.
 *
 *
 * @return 		void
 *********************************************************************
 */
void vApplicationIdleHook (void)
{
	OSALIdleHook();
}

/**
 **********************************************************************
 * @fn			void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed portCHAR *pcTaskName)
 *
 * @brief		This function will be called whenever a task overflowed
 * 				his stack size. The function name is fixed and cannot
 * 				be changed without changing the FreeRTOS source code.
 *
 * @param[in]	pxTask Input 	Pointer to task handle indicating the task which
 * 								generated an stack overflow..
 * @param[in]	pcTaskName		Pointer to the corresponding task name.
 *
 * @return 		void
 *********************************************************************
 */
void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed portCHAR *pcTaskName)
{
	OSALStackOverflowHook((OSAL_TASK*)pxTask, (OSAL_NAME*)pcTaskName);
}

/**
 **********************************************************************
 * @fn			void vApplicationTickHook(void)
 *
 * @brief		System tick hook function for freeRTOS which will be called
 * 				at every system tick interrupt. We just call the
 * 				@ref OSALSysTickHook() function, which has been
 * 				declared as WEAK, thus can be overwritten by the
 * 				application layer.
 *
 *
 * @return 		void
 *********************************************************************
 */
void vApplicationTickHook (void);
void vApplicationTickHook (void)
{
	OSALSysTickHook();
}

OSAL_ERROR OSW_QueueAddToRegistry(OSAL_QUEUE hQueue, OSAL_NAME *pName)
{
	vQueueAddToRegistry(hQueue, (signed char *)pName);
	return OSAL_ERROR_NONE;
}
